#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np
from astropy import coordinates
from astropy import units as u
from astropy import wcs
from astropy.io import fits

from ginga.util.wcsmod import common

try:
    import gwcs  # noqa
    HAVE_GWCS = True
except ImportError:
    HAVE_GWCS = False

try:
    import sunpy.coordinates  # noqa
except ImportError:
    pass
coord_types = [f.name for f in
               coordinates.frame_transform_graph.frame_set]


class AstropyWCS(common.BaseWCS):
    """
    A WCS interface for astropy.wcs as defined by Astropy APE 14.
    You need to install python module 'astropy'
    (http://pypi.python.org/pypi/astropy) v3.1 or later
    if you want to use this version.

    """
    def __init__(self, logger):
        super(AstropyWCS, self).__init__(logger)
        self.kind = 'astropy/APE14'
        self.coordsys = 'raw'  # Remove when load_header() sorted out

        # NOTE: self.wcs magically set in loader right now.

    # TODO: This was copied from wcs_astropy.py. Modify to be compatible
    #       with both FITS WCS and GWCS. Needed so FITS WCS still displays
    #       if user chooses this WCSpkg because APE14 is supposed to be
    #       compatible with both. Nadia will have a PR for GWCS soon.
    def load_header(self, header, fobj=None):
        try:
            # reconstruct a FITS because otherwise we take an
            # incredible performance hit in astropy.wcs
            self.logger.debug("Reconstructing astropy.io.fits header")
            self.header = fits.Header(header.items())

            self.logger.debug("Trying to make astropy-- wcs object")
            self.wcs = wcs.WCS(self.header, fobj=fobj, relax=True)
            self.logger.debug("made astropy wcs object")

            self.coordsys = common.get_coord_system_name(self.header)
            self.logger.debug("Coordinate system is: {}".format(self.coordsys))

        except Exception as e:
            self.logger.error("Error making WCS object: {}".format(str(e)))
            self.wcs = None

    def load_nddata(self, ndd):
        try:
            # reconstruct a pyfits header, because otherwise we take an
            # incredible performance hit in astropy.wcs
            self.logger.debug("Reconstructing astropy.io.fits header")
            self.header = fits.Header(ndd.meta)

            if ndd.wcs is None:
                self.logger.debug("Trying to make astropy FITS WCS object")
                self.wcs = wcs.WCS(self.header, relax=True)
                self.logger.debug("made astropy wcs object")
            else:
                self.logger.debug("reused nddata wcs object")
                self.wcs = ndd.wcs

            if HAVE_GWCS and isinstance(self.wcs, gwcs.WCS):
                self.coordsys = self.wcs.output_frame.name
            else:  # FITS WCS
                self.coordsys = common.get_coord_system_name(self.header)
            self.logger.debug("Coordinate system is: {}".format(self.coordsys))

        except Exception as e:
            self.logger.error("Error making WCS object: {}".format(str(e)))
            self.wcs = None

    def spectral_coord(self, idxs, coords='data'):
        # NOTE: origin is always 0, coords unused.
        pixcrd = np.array([idxs], np.float_)
        try:
            sky = self.wcs.pixel_to_world(
                pixcrd[:, 0], pixcrd[:, 1], pixcrd[:, 2])
            return sky[1].value[0]

        except Exception as e:
            self.logger.error(
                "Error calculating spectral coordinate: {}".format(str(e)))
            raise common.WCSError(e)

    def pixtoradec(self, idxs, coords='data'):
        # NOTE: origin is always 0, coords unused.
        try:
            c = self.wcs.pixel_to_world(*idxs)
            if (isinstance(c, list) and
                    isinstance(c[0], coordinates.SkyCoord)):  # naxis > 2
                c = c[0]
        except Exception as e:
            self.logger.error(
                "Error calculating pixtoradec: {}".format(str(e)))
            raise common.WCSError(e)

        if isinstance(c, coordinates.SkyCoord):
            radec = (c.ra.deg, c.dec.deg)
        else:  # list of Quantity (e.g., from primary header)
            radec = (c[0].value, c[1].value)

        return radec

    def radectopix(self, ra_deg, dec_deg, coords='data', naxispath=None):
        # NOTE: origin is always 0, coords unused.

        args = [ra_deg, dec_deg]
        if naxispath:
            args += [0] * len(naxispath)

        try:
            xy = self.wcs.world_to_pixel_values(*args)[:2]
        except Exception as e:
            self.logger.error(
                "Error calculating radectopix: {}".format(str(e)))
            raise common.WCSError(e)

        return xy

    def pixtocoords(self, idxs, system=None, coords='data'):
        # NOTE: origin is always 0, coords unused.
        if self.coordsys == 'raw':
            raise common.WCSError("No usable WCS")

        if system is None:
            system = 'icrs'

        # Get a coordinates object based on ra/dec wcs transform
        coord = self.wcs.pixel_to_world(*idxs)
        if isinstance(coord, list):  # naxis > 2
            coord = coord[0]
        to_class = coordinates.frame_transform_graph.lookup_name(system)

        # Skip in input and output is the same (no realize_frame
        # call in astropy)
        if to_class != coord.name:
            coord = coord.transform_to(to_class)

        return coord

    def pixtosystem(self, idxs, system=None, coords='data'):
        if self.coordsys == 'pixel':
            return self.pixtoradec(idxs, coords=coords)

        c = self.pixtocoords(idxs, system=system, coords=coords)
        r = c.data.represent_as(coordinates.UnitSphericalRepresentation)
        return r.lon.deg, r.lat.deg

    def datapt_to_wcspt(self, datapt, coords='data', naxispath=None):
        # NOTE: origin is always 0, coords unused.
        if naxispath is not None:
            n = len(naxispath)
            if n > 0:
                datapt = np.hstack((datapt, np.zeros((len(datapt), n))))
        datapt = np.asarray(datapt)
        try:
            args = [datapt[:, i] for i in range(datapt.shape[1])]
            # NOTE: Ignores system transformation.
            wcspt = np.array(self.wcs.pixel_to_world_values(*args)).T
        except Exception as e:
            self.logger.error(
                "Error calculating datapt_to_wcspt: {}".format(str(e)))
            raise common.WCSError(e)

        return wcspt

    def wcspt_to_datapt(self, wcspt, coords='data', naxispath=None):
        # NOTE: origin is always 0, coords unused.
        if naxispath is not None:
            n = len(naxispath)
            if n > 0:
                wcspt = np.hstack((wcspt, np.zeros((len(wcspt), n))))
        wcspt = np.asarray(wcspt)
        try:
            args = [wcspt[:, i] for i in range(wcspt.shape[1])]
            # NOTE: Ignores system transformation.
            datapt = np.asarray(self.wcs.world_to_pixel_values(*args))[:2, :].T
        except Exception as e:
            self.logger.error(
                "Error calculating wcspt_to_datapt: {}".format(str(e)))
            raise common.WCSError(e)

        return datapt

    def datapt_to_system(self, datapt, system=None, coords='data',
                         naxispath=None):
        """
        Map points to given coordinate system.

        Parameters
        ----------
        datapt : array-like
            Pixel coordinates in the format of
            ``[[x0, y0, ...], [x1, y1, ...], ..., [xn, yn, ...]]``.

        system : str or None, optional, default to 'icrs'
            Coordinate system name.

        coords : 'data' or None, optional, default to 'data'
            Expresses whether the data coordinate is indexed from zero

        naxispath : list-like or None, optional, defaults to None
            A sequence defining the pixel indexes > 2D, if any

        Returns
        -------
        coord : SkyCoord

        """
        if self.coordsys == 'raw':
            raise common.WCSError("No usable WCS")
        elif self.coordsys == 'world':
            coordsys = 'icrs'
        else:
            coordsys = self.coordsys

        if system is None:
            system = 'icrs'

        # Get a coordinates object based on ra/dec wcs transform
        wcspt = self.datapt_to_wcspt(datapt, coords=coords,
                                     naxispath=naxispath)
        frame_class = coordinates.frame_transform_graph.lookup_name(coordsys)
        ra_deg = wcspt[:, 0]
        dec_deg = wcspt[:, 1]
        coord = frame_class(ra_deg * u.degree, dec_deg * u.degree)
        to_class = coordinates.frame_transform_graph.lookup_name(system)

        # Skip in input and output is the same (no realize_frame
        # call in astropy)
        if to_class != frame_class:
            coord = coord.transform_to(to_class)

        return coord


# register our WCS with ginga
common.register_wcs('astropy_ape14', AstropyWCS, coord_types)



#
# ModuleManager.py -- Simple class to dynamically manage module imports.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
Dynamically manage module imports.

The ModuleManager class
"""
import sys
import os
import importlib

__all__ = ['ModuleManager']


def my_import(name, path=None):
    """Return imported module for the given name."""

    # Documentation for importlib says this may be needed to pick up
    # modules created after the program has started
    if hasattr(importlib, 'invalidate_caches'):
        # python 3.3+
        importlib.invalidate_caches()

    if path is not None:
        directory, src_file = os.path.split(path)

        # TODO: use the importlib.util machinery
        sys.path.insert(0, directory)
        try:
            mod = importlib.import_module(name)
        finally:
            sys.path.pop(0)

    else:
        mod = importlib.import_module(name)

    return mod


class ModuleManagerError(Exception):
    """Exception related to module manager."""
    pass


class ModuleManager(object):
    """Simple class to dynamically manage module imports."""

    def __init__(self, logger):
        self.logger = logger

        self.module = {}

    def load_module(self, module_name, pfx=None, path=None):
        """Load/reload module from the given name."""
        try:
            if pfx:
                name = pfx + '.' + module_name
            else:
                name = module_name

            if name in sys.modules:
                module = sys.modules[name]
                self.logger.info("Reloading module '%s'..." % module_name)
                module = importlib.reload(module)

            else:
                self.logger.info("Loading module '%s'..." % module_name)
                module = my_import(name, path=path)

            self.module[module_name] = module
            return module

        except Exception as e:
            self.logger.error("Failed to load module '%s': %s" % (
                module_name, str(e)))
            raise ModuleManagerError(e)

    def get_module(self, module_name):
        """Return loaded module from the given name."""
        try:
            return self.module[module_name]

        except KeyError:
            return sys.modules[module_name]

    ########################################################
    ### NON-PEP8 PREDECESSORS: TO BE DEPRECATED

    loadModule = load_module
    getModule = get_module

# END



#
# convenience module for importing all canvas object classes
#

from .basic import *  # noqa
from .image import *  # noqa
from .layer import *  # noqa
from .utils import *  # noqa
from .astro import *  # noqa

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
A plugin for generating a plot of the values along a line or path.

**Plugin Type: Local**

``Cuts`` is a local plugin, which means it is associated with a
channel.  An instance can be opened for each channel.

**Usage**

``Cuts`` plots a simple graph of pixel values vs. index for a line drawn
through the image. Multiple cuts can be plotted.

There are four kinds of cuts available: line, path, freepath and
beziercurve:

* The "line" cut is a straight line between two points.
* The "path" cut is drawn like an open polygon, with straight segments
  in-between.
* The "freepath" cut is like a path cut, but drawn using a free-form
  stroke following the cursor movement.
* The "beziercurve" path is a cubic Bezier curve.

If a new image is added to the channel while the plugin is active, it
will update with the new calculated cuts on the new image.

If the "enable slit" setting is enabled, this plugin will also allow
slit image functionality (for multidimensional images) via a "Slit" tab.
In the tab UI, select one axis from the "Axes" list and draw a line.
This will create a 2D image that assumes the first two axes are
spatial and index the data along the selected axis.
Much like ``Cuts``, you can view the other slit images using the cut
selection drop down box.

**Drawing Cuts**

The "New Cut Type" menu let you choose what kind of cut you are going to draw.

Choose "New Cut" from the "Cut" dropdown menu if you want to draw a
new cut. Otherwise, if a particular named cut is selected then that
will be replaced by any newly drawn cut.

While drawing a path or beziercurve cut, press 'v' to add a vertex,
or 'z' to remove the last vertex added.

**Keyboard Shortcuts**

While hovering the cursor, press 'h' for a full horizontal cut and
'j' for a full vertical cut.

**Deleting Cuts**

To delete a cut, select its name from the "Cut" dropdown and click the
"Delete" button.  To delete all cuts, press "Delete All".

**Editing Cuts**

Using the edit canvas function, it is possible to add new vertices to
an existing path and to move vertices around.   Click the "Edit"
radio button to put the canvas in edit mode.  If a cut is not
automatically selected, you can now select the line, path, or curve by
clicking on it, which should enable the control points at the ends or
vertices -- you can drag these around.  To add a new vertex to a path,
hover the cursor carefully on the line where you want the new vertex
and press 'v'.  To get rid of a vertex, hover the cursor over it and
press 'z'.

You will notice one extra control point for most objects, which has
a center of a different color -- this is a movement control point for
moving the entire object around the image when in edit mode.

You can also select "Move" to just move a cut unchanged.

**Changing Width of Cuts**

The width of 'line' cuts can be changed using the "Width Type" menu:

* "none" indicates a cut of zero radius; i.e., only showing the pixel
  values along the line
* "x" will plot the sum of values along the X axis orthogonal to the cut.
* "y" will plot the sum of values along the Y axis orthogonal to the cut.
* "perpendicular" will plot the sum of values along an axis perpendicular
  to the cut.

The "Width radius" controls the width of the orthogonal summation by
an amount on either side of the cut -- 1 would be 3 pixels, 2 would be 5
pixels, etc.

**Saving Cuts**

Use the "Save" button to save the ``Cuts`` plot as as image and
data as a Numpy compressed archive.

**User Configuration**

"""
import numpy as np

from ginga.gw import Widgets
from ginga import GingaPlugin, colors
from ginga.canvas.coordmap import OffsetMapper

try:
    from ginga.gw import Plot
    from ginga.util import plots
    have_mpl = True
except ImportError:
    have_mpl = False

__all__ = ['Cuts']

# default cut colors
cut_colors = ['magenta', 'skyblue2', 'chartreuse2', 'cyan', 'pink',
              'burlywood2', 'yellow3', 'turquoise', 'coral1', 'mediumpurple2']


class Cuts(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Cuts, self).__init__(fv, fitsimage)

        self.layertag = 'cuts-canvas'
        self._new_cut = 'New Cut'
        self.cutstag = self._new_cut
        self.tags = [self._new_cut]
        self.count = 0
        self.cuttypes = ['line', 'path', 'freepath', 'beziercurve']
        self.cuttype = 'line'
        self.save_enabled = False
        # for 3D Slit functionality
        self.transpose_enabled = False
        self.selected_axis = None
        self.hbox_axes = None
        self._split_sizes = [400, 500]

        # For collecting data orthogonal to the cut
        self.widthtypes = ['none', 'x', 'y', 'perpendicular']
        self.widthtype = 'none'
        self.width_radius = 5
        self.tine_spacing_px = 100

        # get Cuts preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Cuts')
        self.settings.add_defaults(select_new_cut=True, draw_then_move=True,
                                   label_cuts=True, colors=cut_colors,
                                   drag_update=False,
                                   show_cuts_legend=False, enable_slit=False)
        self.settings.load(onError='silent')
        self.colors = self.settings.get('colors', cut_colors)

        self.dc = fv.get_draw_classes()
        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype('line', color='cyan', linestyle='dash')
        canvas.set_callback('draw-event', self.draw_cb)
        canvas.set_callback('edit-event', self.edit_cb)
        canvas.add_draw_mode('move', down=self.buttondown_cb,
                             move=self.motion_cb, up=self.buttonup_cb,
                             key=self.keydown)
        canvas.set_draw_mode('draw')
        canvas.register_for_cursor_drawing(self.fitsimage)
        canvas.set_surface(self.fitsimage)
        self.canvas = canvas

        self.use_slit = self.settings.get('enable_slit', False)
        self.cuts_image = None

        self.gui_up = False

    def build_gui(self, container):
        if not have_mpl:
            raise ImportError('Install matplotlib to use this plugin')

        top = Widgets.VBox()
        top.set_border_width(4)

        # Make the cuts plot
        box, sw, orientation = Widgets.get_oriented_box(container)
        box.set_margins(4, 4, 4, 4)
        box.set_spacing(2)

        paned = Widgets.Splitter(orientation=orientation)
        self.w.splitter = paned

        # Add Tab Widget
        nb = Widgets.TabWidget(tabpos='top')
        paned.add_widget(Widgets.hadjust(nb, orientation))

        self.cuts_plot = plots.CutsPlot(logger=self.logger,
                                        width=400, height=400)
        self.plot = Plot.PlotWidget(self.cuts_plot)
        self.plot.resize(400, 400)
        ax = self.cuts_plot.add_axis()
        ax.grid(True)

        self.slit_plot = plots.Plot(logger=self.logger,
                                    width=400, height=400)
        if plots.MPL_GE_2_0:
            kwargs = {'facecolor': 'black'}
        else:
            kwargs = {'axisbg': 'black'}
        self.slit_plot.add_axis(**kwargs)
        self.plot2 = Plot.PlotWidget(self.slit_plot)
        self.plot2.resize(400, 400)

        captions = (('Cut:', 'label', 'Cut', 'combobox',
                     'New Cut Type:', 'label', 'Cut Type', 'combobox'),
                    ('Delete Cut', 'button', 'Delete All', 'button'),
                    ('Save', 'button'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        # control for selecting a cut
        combobox = b.cut
        for tag in self.tags:
            combobox.append_text(tag)
        combobox.show_text(self.cutstag)
        combobox.add_callback('activated', self.cut_select_cb)
        self.w.cuts = combobox
        combobox.set_tooltip("Select a cut to redraw or delete")

        # control for selecting cut type
        combobox = b.cut_type
        for cuttype in self.cuttypes:
            combobox.append_text(cuttype)
        self.w.cuts_type = combobox
        index = self.cuttypes.index(self.cuttype)
        combobox.set_index(index)
        combobox.add_callback('activated', self.set_cutsdrawtype_cb)
        combobox.set_tooltip("Choose the cut type to draw")

        self.save_cuts = b.save
        self.save_cuts.set_tooltip("Save cuts plot and data")
        self.save_cuts.add_callback('activated',
                                    lambda w: self.save_cb(mode='cuts'))
        self.save_cuts.set_enabled(self.save_enabled)

        btn = b.delete_cut
        btn.add_callback('activated', self.delete_cut_cb)
        btn.set_tooltip("Delete selected cut")

        btn = b.delete_all
        btn.add_callback('activated', self.delete_all_cb)
        btn.set_tooltip("Clear all cuts")

        fr = Widgets.Frame("Cuts")
        fr.set_widget(w)

        box.add_widget(fr, stretch=0)

        exp = Widgets.Expander("Cut Width")

        captions = (('Width Type:', 'label', 'Width Type', 'combobox',
                     'Width radius:', 'label', 'Width radius', 'spinbutton'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        # control for selecting width cut type
        combobox = b.width_type
        for atype in self.widthtypes:
            combobox.append_text(atype)
        index = self.widthtypes.index(self.widthtype)
        combobox.set_index(index)
        combobox.add_callback('activated', self.set_width_type_cb)
        combobox.set_tooltip("Direction of summation orthogonal to cut")

        sb = b.width_radius
        sb.add_callback('value-changed', self.width_radius_changed_cb)
        sb.set_tooltip("Radius of cut width")
        sb.set_limits(1, 100)
        sb.set_value(self.width_radius)

        fr = Widgets.Frame()
        fr.set_widget(w)
        exp.set_widget(fr)

        box.add_widget(exp, stretch=0)
        box.add_widget(Widgets.Label(''), stretch=1)
        paned.add_widget(sw)
        paned.set_sizes(self._split_sizes)

        top.add_widget(paned, stretch=5)

        mode = self.canvas.get_draw_mode()
        hbox = Widgets.HBox()
        btn1 = Widgets.RadioButton("Move")
        btn1.set_state(mode == 'move')
        btn1.add_callback('activated',
                          lambda w, val: self.set_mode_cb('move', val))
        btn1.set_tooltip("Choose this to position cuts")
        self.w.btn_move = btn1
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton("Draw", group=btn1)
        btn2.set_state(mode == 'draw')
        btn2.add_callback('activated',
                          lambda w, val: self.set_mode_cb('draw', val))
        btn2.set_tooltip("Choose this to draw a new or replacement cut")
        self.w.btn_draw = btn2
        hbox.add_widget(btn2)

        btn3 = Widgets.RadioButton("Edit", group=btn1)
        btn3.set_state(mode == 'edit')
        btn3.add_callback('activated',
                          lambda w, val: self.set_mode_cb('edit', val))
        btn3.set_tooltip("Choose this to edit a cut")
        self.w.btn_edit = btn3
        hbox.add_widget(btn3)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(hbox, stretch=0)

        # Add Cuts plot to its tab
        vbox_cuts = Widgets.VBox()
        vbox_cuts.add_widget(self.plot, stretch=1)
        nb.add_widget(vbox_cuts, title="Cuts")

        captions = (("Enable Slit", 'checkbutton',
                     "Transpose Plot", 'checkbutton', "Save", 'button'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        def chg_enable_slit(w, val):
            self.use_slit = val
            if val:
                self.build_axes()
            return True
        b.enable_slit.set_state(self.use_slit)
        b.enable_slit.set_tooltip("Enable the slit function")
        b.enable_slit.add_callback('activated', chg_enable_slit)

        self.t_btn = b.transpose_plot
        self.t_btn.set_tooltip("Flip the plot")
        self.t_btn.set_state(self.transpose_enabled)
        self.t_btn.add_callback('activated', self.transpose_plot)

        self.save_slit = b.save
        self.save_slit.set_tooltip("Save slit plot and data")
        self.save_slit.add_callback('activated',
                                    lambda w: self.save_cb(mode='slit'))
        self.save_slit.set_enabled(self.save_enabled)

        # Add frame to hold the slit controls
        fr = Widgets.Frame("Axes controls")
        self.hbox_axes = Widgets.HBox()
        self.hbox_axes.set_border_width(4)
        self.hbox_axes.set_spacing(1)
        fr.set_widget(self.hbox_axes)

        # Add Slit plot and controls to its tab
        vbox_slit = Widgets.VBox()
        vbox_slit.add_widget(self.plot2, stretch=1)
        vbox_slit.add_widget(w)
        vbox_slit.add_widget(fr)
        nb.add_widget(vbox_slit, title="Slit")

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)

        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)

        self.select_cut(self.cutstag)
        self.gui_up = True

        if self.use_slit:
            self.build_axes()

    def build_axes(self):
        self.selected_axis = None
        if (not self.gui_up) or (self.hbox_axes is None):
            return
        self.hbox_axes.remove_all()
        image = self.fitsimage.get_image()
        if image is not None:
            # Add Checkbox widgets
            # `image.naxispath` returns only mdim axes
            for i in range(1, len(image.naxispath) + 3):
                chkbox = Widgets.CheckBox('NAXIS%d' % i)
                self.hbox_axes.add_widget(chkbox)

                # Disable axes 1,2
                if i < 3:
                    chkbox.set_enabled(False)
                    continue

                # Add callback
                self.axes_callback_handler(chkbox, i)

            self.redraw_slit('clear')

    def axes_callback_handler(self, chkbox, pos):
        chkbox.add_callback('activated',
                            lambda w, tf: self.axis_toggle_cb(w, tf, pos))

    def select_cut(self, tag):
        self.cutstag = tag
        self.w.cuts.show_text(tag)

        if (tag == self._new_cut) or len(self.tags) < 2:
            self.w.delete_cut.set_enabled(False)

            self.w.btn_move.set_enabled(False)
            self.w.btn_edit.set_enabled(False)
            self.set_mode('draw')

            if self.use_slit:
                self.redraw_slit('clear')
        else:
            self.w.delete_cut.set_enabled(True)

            self.w.btn_move.set_enabled(True)
            self.w.btn_edit.set_enabled(True)

            if self.w.btn_edit.get_state():
                self.edit_select_cuts()

            if self.use_slit:
                self._plot_slit()

    def cut_select_cb(self, w, index):
        tag = self.tags[index]
        self.select_cut(tag)

    def set_cutsdrawtype_cb(self, w, index):
        self.cuttype = self.cuttypes[index]
        self.canvas.set_drawtype(self.cuttype, color='cyan', linestyle='dash')

    def delete_cut_cb(self, w):
        tag = self.cutstag
        if tag == self._new_cut:
            return
        index = self.tags.index(tag)  # noqa
        self.canvas.delete_object_by_tag(tag)
        self.w.cuts.delete_alpha(tag)
        self.tags.remove(tag)
        idx = len(self.tags) - 1
        tag = self.tags[idx]
        self.select_cut(tag)
        if tag == self._new_cut:
            self.save_cuts.set_enabled(False)
            if self.use_slit and self.gui_up:
                self.save_slit.set_enabled(False)
        # plot cleared in replot_all() if no more cuts
        self.replot_all()

    def delete_all_cb(self, w):
        self.canvas.delete_all_objects()
        self.w.cuts.clear()
        self.tags = [self._new_cut]
        self.cutstag = self._new_cut
        self.w.cuts.append_text(self._new_cut)
        self.select_cut(self._new_cut)
        self.save_cuts.set_enabled(False)
        if self.use_slit and self.gui_up:
            self.save_slit.set_enabled(False)
        # plot cleared in replot_all() if no more cuts
        self.replot_all()

    def add_cuts_tag(self, tag):
        if tag not in self.tags:
            self.tags.append(tag)
            self.w.cuts.append_text(tag)

        select_flag = self.settings.get('select_new_cut', True)
        if select_flag:
            self.select_cut(tag)
            move_flag = self.settings.get('draw_then_move', True)
            if move_flag:
                self.set_mode('move')

    def close(self):
        #self.set_mode('move')
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        # start line cuts operation
        self.cuts_plot.set_titles(rtitle="Cuts")

        self.drag_update = self.settings.get('drag_update', False)

        # insert canvas, if not already
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.get_object_by_tag(self.layertag)

        except KeyError:
            # Add ruler layer
            p_canvas.add(self.canvas, tag=self.layertag)

        #self.canvas.delete_all_objects()
        self.resume()

    def pause(self):
        self.canvas.ui_set_active(False)

    def resume(self):
        # turn off any mode user may be in
        self.modes_off()

        self.canvas.ui_set_active(True)
        self.fv.show_status("Draw a line with the right mouse button")
        self.replot_all()
        if self.use_slit:
            self.cuts_image = self.fitsimage.get_image()

    def stop(self):
        self.gui_up = False
        self._split_sizes = self.w.splitter.get_sizes()
        # remove the canvas from the image
        p_canvas = self.fitsimage.get_canvas()
        p_canvas.delete_object_by_tag(self.layertag)
        self.fv.show_status("")

    def redo(self):
        """This is called when a new image arrives or the data in the
        existing image changes.
        """
        if self.use_slit:
            image = self.fitsimage.get_image()
            if image != self.cuts_image:
                self.cuts_image = image
                self.build_axes()

        self.replot_all()

    def _get_perpendicular_points(self, obj, x, y, r):
        dx = float(obj.x1 - obj.x2)
        dy = float(obj.y1 - obj.y2)
        dist = np.sqrt(dx * dx + dy * dy)
        dx /= dist
        dy /= dist
        x3 = x + r * dy
        y3 = y - r * dx
        x4 = x - r * dy
        y4 = y + r * dx
        return (x3, y3, x4, y4)

    def _get_width_points(self, obj, x, y, rx, ry):
        x3, y3 = x - rx, y - ry
        x4, y4 = x + rx, y + ry
        return (x3, y3, x4, y4)

    def get_orthogonal_points(self, obj, x, y, r):
        if self.widthtype == 'x':
            return self._get_width_points(obj, x, y, r, 0)
        elif self.widthtype == 'y':
            return self._get_width_points(obj, x, y, 0, r)
        else:
            return self._get_perpendicular_points(obj, x, y, r)

    def get_orthogonal_array(self, image, obj, x, y, r):
        x1, y1, x2, y2 = self.get_orthogonal_points(obj, x, y, r)
        values = image.get_pixels_on_line(int(x1), int(y1),
                                          int(x2), int(y2))
        return np.array(values)

    def _plotpoints(self, obj, color):

        image = self.fitsimage.get_image()

        # Get points on the line
        if obj.kind == 'line':
            if self.widthtype == 'none':
                points = image.get_pixels_on_line(int(obj.x1), int(obj.y1),
                                                  int(obj.x2), int(obj.y2))
            else:
                coords = image.get_pixels_on_line(int(obj.x1), int(obj.y1),
                                                  int(obj.x2), int(obj.y2),
                                                  getvalues=False)

                points = []
                for x, y in coords:
                    arr = self.get_orthogonal_array(image, obj, x, y,
                                                    self.width_radius)
                    val = np.nansum(arr)
                    points.append(val)

        elif obj.kind in ('path', 'freepath'):
            points = []
            x1, y1 = obj.points[0]
            for x2, y2 in obj.points[1:]:
                pts = image.get_pixels_on_line(int(x1), int(y1),
                                               int(x2), int(y2))
                # don't repeat last point when adding next segment
                points.extend(pts[:-1])
                x1, y1 = x2, y2

        elif obj.kind == 'beziercurve':
            points = obj.get_pixels_on_curve(image)

        points = np.array(points)

        rgb = colors.lookup_color(color)
        self.cuts_plot.cuts(points, xtitle="Line Index", ytitle="Pixel Value",
                            color=rgb)

        if self.settings.get('show_cuts_legend', False):
            self.add_legend()

    def add_legend(self):
        """Add or update Cuts plot legend."""
        cuts = [tag for tag in self.tags if tag is not self._new_cut]
        self.cuts_plot.ax.legend(cuts, loc='best',
                                 shadow=True, fancybox=True,
                                 prop={'size': 8}, labelspacing=0.2)

    def get_coords(self, obj):
        image = self.fitsimage.get_image()

        # Check whether multidimensional
        if len(image.naxispath) <= 0:
            return

        # Get points on the line
        if obj.kind == 'line':
            coords = image.get_pixels_on_line(int(obj.x1), int(obj.y1),
                                              int(obj.x2), int(obj.y2),
                                              getvalues=False)

        elif obj.kind in ('path', 'freepath'):
            coords = []
            x1, y1 = obj.points[0]
            for x2, y2 in obj.points[1:]:
                pts = image.get_pixels_on_line(int(x1), int(y1),
                                               int(x2), int(y2),
                                               getvalues=False)
                # don't repeat last point when adding next segment
                coords.extend(pts[:-1])
                x1, y1 = x2, y2

        elif obj.kind == 'beziercurve':
            coords = obj.get_pixels_on_curve(image, getvalues=False)
            # Exclude NaNs
            coords = [c for c in coords if not np.any(np.isnan(c))]

        shape = image.shape
        # Exclude points outside boundaries
        coords = [(coord[0], coord[1]) for coord in coords
                  if (0 <= coord[0] < shape[1] and 0 <= coord[1] < shape[0])]
        if not coords:
            self.redraw_slit('clear')
            return

        return np.array(coords)

    def get_slit_data(self, coords):
        image = self.fitsimage.get_image()
        data = image.get_mddata()
        naxes = data.ndim

        # Small correction
        selected_axis = abs(self.selected_axis - naxes)

        spatial_axes = [naxes - 1, naxes - 2]

        # Build N-dim slice
        axes_slice = image.revnaxis + [0, 0]

        # Slice data according to axis
        for i, sa in enumerate(spatial_axes):
            axes_slice[sa] = coords[:, i]
        axes_slice[selected_axis] = slice(None, None, None)

        self.slit_data = data[tuple(axes_slice)]

    def _plot_slit(self):
        if not self.selected_axis:
            return

        obj = self.canvas.get_object_by_tag(self.cutstag)
        line = self._getlines(obj)

        coords = self.get_coords(line[0])
        self.get_slit_data(coords)

        if self.transpose_enabled:
            self.redraw_slit('transpose')

        else:
            self.slit_plot.ax.imshow(
                self.slit_data, interpolation='nearest',
                origin='lower', aspect='auto').set_cmap('gray')
            self.set_labels()
            self.slit_plot.draw()

    def _replot(self, lines, colors):
        for idx in range(len(lines)):
            line, color = lines[idx], colors[idx]
            line.color = color
            self._plotpoints(line, color)

        return True

    def replot_all(self):
        self.cuts_plot.clear()
        self.w.delete_all.set_enabled(False)
        self.save_cuts.set_enabled(False)
        if self.use_slit and self.gui_up:
            self.save_slit.set_enabled(False)

        idx = 0
        for cutstag in self.tags:
            if cutstag == self._new_cut:
                continue
            obj = self.canvas.get_object_by_tag(cutstag)
            if obj.kind != 'compound':
                continue
            lines = self._getlines(obj)
            n = len(lines)
            count = obj.get_data('count', self.count)
            idx = (count + n) % len(self.colors)
            colors = self.colors[idx:idx + n]
            # text should take same color as first line in line set
            text = obj.objects[1]
            if text.kind == 'text':
                text.color = colors[0]
            #text.color = color
            self._replot(lines, colors)
            self.save_cuts.set_enabled(True)
            self.w.delete_all.set_enabled(True)

        # Redraw slit image for selected cut
        if self.use_slit:
            if self.cutstag != self._new_cut:
                self._plot_slit()
                if self.selected_axis and self.gui_up:
                    self.save_slit.set_enabled(True)

        # force mpl redraw
        self.cuts_plot.draw()

        self.canvas.redraw(whence=3)
        self.fv.show_status(
            "Click or drag left mouse button to reposition cuts")
        return True

    def _create_cut(self, x, y, count, x1, y1, x2, y2, color='cyan'):
        text = "cuts%d" % (count)
        if not self.settings.get('label_cuts', False):
            text = ''
        line_obj = self.dc.Line(x1, y1, x2, y2, color=color,
                                showcap=False)
        text_obj = self.dc.Text(0, 0, text, color=color, coord='offset',
                                ref_obj=line_obj)
        obj = self.dc.CompoundObject(line_obj, text_obj)
        # this is necessary for drawing cuts with width feature
        obj.initialize(self.canvas, self.fitsimage, self.logger)
        obj.set_data(cuts=True)
        return obj

    def _update_tines(self, obj):
        if obj.objects[0].kind != 'line':
            # right now we only know how to adjust lines
            return

        # Remove previous tines, if any
        if len(obj.objects) > 2:
            obj.objects = obj.objects[:2]

        if self.widthtype == 'none':
            return

        image = self.fitsimage.get_image()
        line = obj.objects[0]
        coords = image.get_pixels_on_line(int(line.x1), int(line.y1),
                                          int(line.x2), int(line.y2),
                                          getvalues=False)
        crdmap = OffsetMapper(self.fitsimage, line)
        num_ticks = max(len(coords) // self.tine_spacing_px, 3)
        interval = max(1, len(coords) // num_ticks)
        for i in range(0, len(coords), interval):
            x, y = coords[i]
            x1, y1, x2, y2 = self.get_orthogonal_points(line, x, y,
                                                        self.width_radius)
            (x1, y1), (x2, y2) = crdmap.calc_offsets([(x1, y1), (x2, y2)])
            aline = self.dc.Line(x1, y1, x2, y2)
            aline.crdmap = crdmap
            aline.editable = False
            obj.objects.append(aline)

    def _create_cut_obj(self, count, cuts_obj, color='cyan'):
        text = "cuts%d" % (count)
        if not self.settings.get('label_cuts', False):
            text = ''
        cuts_obj.showcap = False
        cuts_obj.linestyle = 'solid'
        #cuts_obj.color = color
        color = cuts_obj.color
        args = [cuts_obj]
        text_obj = self.dc.Text(0, 0, text, color=color, coord='offset',
                                ref_obj=cuts_obj)
        args.append(text_obj)

        obj = self.dc.CompoundObject(*args)
        obj.set_data(cuts=True)

        if (self.widthtype != 'none') and (self.width_radius > 0):
            self._update_tines(obj)
        return obj

    def _combine_cuts(self, *args):
        return self.dc.CompoundObject(*args)

    def _append_lists(self, l):
        if len(l) == 0:
            return []
        elif len(l) == 1:
            return l[0]
        else:
            res = l[0]
            res.extend(self._append_lists(l[1:]))
            return res

    def _getlines(self, obj):
        if obj.kind == 'compound':
            #return self._append_lists(list(map(self._getlines, obj.objects)))
            return [obj.objects[0]]
        elif obj.kind in self.cuttypes:
            return [obj]
        else:
            return []

    def buttondown_cb(self, canvas, event, data_x, data_y, viewer):
        return self.motion_cb(canvas, event, data_x, data_y, viewer)

    def motion_cb(self, canvas, event, data_x, data_y, viewer):
        if self.cutstag == self._new_cut:
            return True
        obj = self.canvas.get_object_by_tag(self.cutstag)
        # Assume first element of this compound object is the reference obj
        obj = obj.objects[0]
        obj.move_to(data_x, data_y)
        canvas.redraw(whence=3)

        if self.drag_update:
            self.replot_all()
        return True

    def buttonup_cb(self, canvas, event, data_x, data_y, viewer):
        if self.cutstag == self._new_cut:
            return True
        obj = self.canvas.get_object_by_tag(self.cutstag)
        # Assume first element of this compound object is the reference obj
        obj = obj.objects[0]
        obj.move_to(data_x, data_y)

        self.replot_all()
        return True

    def keydown(self, canvas, event, data_x, data_y, viewer):
        if event.key == 'n':
            self.select_cut(self._new_cut)
            return True
        elif event.key == 'h':
            self.cut_at('horizontal')
            return True
        elif event.key == 'j':
            self.cut_at('vertical')
            return True
        return False

    def _get_new_count(self):
        counts = set([])
        for cutstag in self.tags:
            try:
                obj = self.canvas.get_object_by_tag(cutstag)
            except KeyError:
                continue
            counts.add(obj.get_data('count', 0))
        ncounts = set(range(len(self.colors)))
        avail = list(ncounts.difference(counts))
        avail.sort()
        if len(avail) > 0:
            count = avail[0]
        else:
            self.count += 1
            count = self.count
        return count

    def _get_cut_index(self):
        if self.cutstag != self._new_cut:
            # Replacing a cut
            self.logger.debug("replacing cut position")
            try:
                cutobj = self.canvas.get_object_by_tag(self.cutstag)
                self.canvas.delete_object_by_tag(self.cutstag)
                count = cutobj.get_data('count')
            except KeyError:
                count = self._get_new_count()
        else:
            self.logger.debug("adding cut position")
            count = self._get_new_count()
        return count

    def cut_at(self, cuttype):
        """Perform a cut at the last mouse position in the image.
        `cuttype` determines the type of cut made.
        """
        data_x, data_y = self.fitsimage.get_last_data_xy()
        image = self.fitsimage.get_image()
        wd, ht = image.get_size()

        coords = []
        if cuttype == 'horizontal':
            coords.append((0, data_y, wd, data_y))
        elif cuttype == 'vertical':
            coords.append((data_x, 0, data_x, ht))

        count = self._get_cut_index()
        tag = "cuts%d" % (count)
        cuts = []
        for (x1, y1, x2, y2) in coords:
            # calculate center of line
            wd = x2 - x1
            dw = wd // 2
            ht = y2 - y1
            dh = ht // 2
            x, y = x1 + dw + 4, y1 + dh + 4

            cut = self._create_cut(x, y, count, x1, y1, x2, y2, color='cyan')
            self._update_tines(cut)
            cuts.append(cut)

        if len(cuts) == 1:
            cut = cuts[0]
        else:
            cut = self._combine_cuts(*cuts)

        cut.set_data(count=count)

        self.canvas.delete_object_by_tag(tag)
        self.canvas.add(cut, tag=tag)
        self.add_cuts_tag(tag)

        self.logger.debug("redoing cut plots")
        return self.replot_all()

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        canvas.delete_object_by_tag(tag)

        if obj.kind not in self.cuttypes:
            return True

        count = self._get_cut_index()
        tag = "cuts%d" % (count)

        cut = self._create_cut_obj(count, obj, color='cyan')
        cut.set_data(count=count)
        self._update_tines(cut)

        canvas.delete_object_by_tag(tag)
        self.canvas.add(cut, tag=tag)
        self.add_cuts_tag(tag)

        self.logger.debug("redoing cut plots")
        return self.replot_all()

    def edit_cb(self, canvas, obj):
        self.redraw_cuts()
        self.replot_all()
        return True

    def edit_select_cuts(self):
        if self.cutstag != self._new_cut:
            obj = self.canvas.get_object_by_tag(self.cutstag)
            # drill down to reference shape
            if hasattr(obj, 'objects'):
                obj = obj.objects[0]
            self.canvas.edit_select(obj)
        else:
            self.canvas.clear_selected()
        self.canvas.update_canvas()

    def set_mode_cb(self, mode, tf):
        """Called when one of the Move/Draw/Edit radio buttons is selected."""
        if tf:
            self.canvas.set_draw_mode(mode)
            if mode == 'edit':
                self.edit_select_cuts()
        return True

    def set_mode(self, mode):
        self.canvas.set_draw_mode(mode)
        self.w.btn_move.set_state(mode == 'move')
        self.w.btn_draw.set_state(mode == 'draw')
        self.w.btn_edit.set_state(mode == 'edit')

    def redraw_cuts(self):
        """Redraws cuts with tines (for cuts with a 'width')."""
        self.logger.debug("redrawing cuts")
        for cutstag in self.tags:
            if cutstag == self._new_cut:
                continue
            obj = self.canvas.get_object_by_tag(cutstag)
            if obj.kind != 'compound':
                continue
            self._update_tines(obj)
        self.canvas.redraw(whence=3)

    def width_radius_changed_cb(self, widget, val):
        """Callback executed when the Width radius is changed."""
        self.width_radius = val
        self.redraw_cuts()
        self.replot_all()
        return True

    def set_width_type_cb(self, widget, idx):
        self.widthtype = self.widthtypes[idx]
        self.redraw_cuts()
        self.replot_all()
        return True

    def save_cb(self, mode):
        """Save image, figure, and plot data arrays."""

        # This just defines the basename.
        # Extension has to be explicitly defined or things can get messy.
        w = Widgets.SaveDialog(title='Save {0} data'.format(mode))
        filename = w.get_path()

        if filename is None:
            # user canceled dialog
            return

        # TODO: This can be a user preference?
        fig_dpi = 100

        if mode == 'cuts':
            fig, xarr, yarr = self.cuts_plot.get_data()

        elif mode == 'slit':
            fig, xarr, yarr = self.slit_plot.get_data()

        figname = filename + '.png'
        self.logger.info("saving figure as: %s" % (figname))
        fig.savefig(figname, dpi=fig_dpi)

        dataname = filename + '.npz'
        self.logger.info("saving data as: %s" % (dataname))
        np.savez_compressed(dataname, x=xarr, y=yarr)

    def axis_toggle_cb(self, w, tf, pos):
        children = self.hbox_axes.get_children()

        # Deactivate previously selected axis
        if self.selected_axis is not None:
            children[self.selected_axis - 1].set_state(False)

        # Check if the old axis is clicked
        if pos == self.selected_axis:
            self.selected_axis = None
            if self.gui_up:
                self.save_slit.set_enabled(False)
            self.redraw_slit('clear')
        else:
            self.selected_axis = pos
            children[pos - 1].set_state(tf)
            if self.gui_up:
                if self.cutstag != self._new_cut:
                    self.save_slit.set_enabled(True)
                    self._plot_slit()
                else:
                    # no cut selected ("new cut")
                    self.redraw_slit('clear')

    def redraw_slit(self, mode):
        if mode == 'clear':
            self.slit_plot.clear()
        elif mode == 'transpose':
            self.slit_data = self.slit_data.T
            self.slit_plot.ax.imshow(
                self.slit_data, interpolation='nearest',
                origin='lower', aspect='auto').set_cmap('gray')
            self.set_labels()

        self.slit_plot.draw()

    def transpose_plot(self, w, tf):
        old_val = self.transpose_enabled
        self.transpose_enabled = tf

        if (old_val != tf and self.cutstag != self._new_cut and
                self.selected_axis):
            self.redraw_slit('transpose')

    def set_labels(self):
        image = self.fitsimage.get_image()
        shape = image.get_mddata().shape

        if (shape[0] == len(self.slit_data[0]) or
                shape[1] == len(self.slit_data[0])):
            self.slit_plot.ax.set_xlabel('')
            self.slit_plot.ax.set_ylabel('Position along slit')
        else:
            self.slit_plot.ax.set_ylabel('')
            self.slit_plot.ax.set_xlabel('Position along slit')

    def __str__(self):
        return 'cuts'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Cuts', package='ginga')

# END



#
# PluginManager.py -- Simple class to manage plugins.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import threading
import traceback

from ginga.gw import Widgets
from ginga.misc import Bunch, Callback


class PluginManagerError(Exception):
    pass


class PluginManager(Callback.Callbacks):
    """
    A PluginManager manages the start and stop of plugins.
    """

    def __init__(self, logger, gshell, ds, mm):
        super(PluginManager, self).__init__()

        self.logger = logger
        self.fv = gshell
        self.ds = ds
        self.mm = mm

        self.lock = threading.RLock()
        self.plugin = Bunch.caselessDict()
        self.active = {}
        self.focus = set([])
        self.exclusive = set([])

        for name in ('activate-plugin', 'deactivate-plugin',
                     'focus-plugin', 'unfocus-plugin'):
            self.enable_callback(name)

    def load_plugin(self, name, spec, chinfo=None):
        try:
            module = self.mm.get_module(spec.module)
            className = spec.get('klass', spec.module)
            klass = getattr(module, className)

            if chinfo is None:
                # global plug in
                obj = klass(self.fv)
                fitsimage = None
            else:
                # local plugin
                fitsimage = chinfo.fitsimage
                obj = klass(self.fv, fitsimage)

            # Prepare configuration for module.  This becomes the p_info
            # object referred to in later code.
            opname = name.lower()
            self.plugin[opname] = Bunch.Bunch(klass=klass, obj=obj,
                                              widget=None, name=name,
                                              is_toplevel=False,
                                              spec=spec,
                                              fitsimage=fitsimage,
                                              chinfo=chinfo)

            self.logger.info("Plugin '%s' loaded." % name)

        except Exception as e:
            self.logger.error("Failed to load plugin '%s': %s" % (
                name, str(e)))
            try:
                (type, value, tb) = sys.exc_info()
                tb_str = "\n".join(traceback.format_tb(tb))
                self.logger.error("Traceback:\n%s" % (tb_str))

            except Exception as e:
                tb_str = "Traceback information unavailable."
                self.logger.error(tb_str)
            #raise PluginManagerError(e)

    def reload_plugin(self, plname, chinfo=None):
        p_info = self.get_plugin_info(plname)
        return self.load_plugin(p_info.name, p_info.spec, chinfo=chinfo)

    def has_plugin(self, plname):
        plname = plname.lower()
        return plname in self.plugin

    def get_plugin_info(self, plname):
        plname = plname.lower()
        p_info = self.plugin[plname]
        return p_info

    def get_plugin(self, name):
        p_info = self.get_plugin_info(name)
        return p_info.obj

    def get_names(self):
        return list(self.plugin.keys())

    def deactivate_focused(self):
        names = self.get_focus()
        for name in names:
            self.deactivate(name)

    def get_active(self):
        return list(self.active.keys())

    def is_active(self, key):
        lname = key.lower()
        return lname in self.get_active()

    def get_focus(self):
        return list(self.focus)

    def has_focus(self, name):
        lname = name.lower()
        names = self.get_focus()
        return lname in names

    def get_info(self, name):
        lname = name.lower()
        return self.active[lname]

    def activate(self, p_info, exclusive=True):
        name = p_info.tabname
        lname = p_info.name.lower()
        if lname in self.active:
            # plugin already active
            return

        bnch = Bunch.Bunch(pInfo=p_info, lblname=name, widget=None,
                           exclusive=exclusive)

        if p_info.chinfo is not None:
            # local plugin
            tup = name.split(':')
            bnch.lblname = ' ' + tup[0] + ':\n' + tup[1] + ' '
        else:
            # global plugin
            bnch.exclusive = False

        self.active[lname] = bnch
        if bnch.exclusive:
            self.exclusive.add(lname)

        self.make_callback('activate-plugin', bnch)

    def deactivate(self, name):
        self.logger.debug("deactivating %s" % (name))
        lname = name.lower()
        if lname in self.focus:
            self.clear_focus(lname)

        if lname not in self.active:
            # plugin already deactivated
            return

        self.logger.debug("removing from task bar: %s" % (lname))
        bnch = self.active[lname]

        del self.active[lname]

        try:
            self.stop_plugin(bnch.pInfo)

        except Exception as e:
            self.logger.error("Error deactivating plugin: %s" % (str(e)))
            # TODO: log traceback

        # Set focus to another plugin if one is running, but only if it
        # is a local plugin
        if bnch.pInfo.spec.ptype != 'global':
            active = list(self.active.keys())
            if len(active) > 0:
                name = active[0]
                self.logger.debug("focusing: %s" % (name))
                self.set_focus(name)

        self.make_callback('deactivate-plugin', bnch)

    def stop_reload_start(self, name):
        # deactivate the running plugin, if active
        if self.is_active(name):
            self.deactivate(name)

        self.fv.update_pending(timeout=0.1)

        # reload the plugin
        p_info = self.get_plugin_info(name)
        chname = None
        if p_info.chinfo is not None:
            chname = p_info.chinfo.name

        self.reload_plugin(name, chinfo=p_info.chinfo)

        # and start it up again
        self.start_plugin_future(chname, name, None)

    def set_focus(self, name):
        self.logger.debug("Focusing plugin '%s'" % (name))
        lname = name.lower()
        bnch = self.active[lname]
        if bnch.exclusive:
            self.logger.debug("focus=%s exclusive=%s" % (
                self.focus, self.exclusive))
            defocus = list(filter(lambda x: x in self.exclusive, self.focus))
            self.logger.debug("defocus: %s" % (str(defocus)))
            for xname in defocus:
                self.clear_focus(xname)

        p_info = bnch.pInfo
        # If this is a local plugin, raise the channel associated with the
        # plug in
        if p_info.chinfo is not None:
            self.logger.debug("resuming plugin %s" % (name))
            p_info.obj.resume()

        self.focus.add(lname)
        self.make_callback('focus-plugin', bnch)

        if p_info.widget is not None:
            self.logger.debug("raising plugin tab %s" % (p_info.tabname))
            if p_info.is_toplevel:
                p_info.widget.raise_()
            else:
                self.ds.raise_tab(p_info.tabname)

        if p_info.chinfo is not None:
            chname = p_info.chinfo.name
            self.logger.debug("changing channel to %s" % (chname))
            # Alternative could just be to raise the channel tab rather
            # than making all the global plugins switch over to the new
            # channel
            #self.ds.raise_tab(chname)
            self.fv.change_channel(chname)

    def clear_focus(self, name):
        self.logger.debug("Unfocusing plugin '%s'" % (name))
        lname = name.lower()
        bnch = self.active[lname]
        p_info = bnch.pInfo
        try:
            self.focus.remove(lname)

            if p_info.chinfo is not None:
                p_info.obj.pause()

            self.make_callback('unfocus-plugin', bnch)

        except Exception as e:
            self.logger.error("Error unfocusing plugin '%s': %s" % (
                name, str(e)))

    def start_plugin(self, chname, opname, alreadyOpenOk=False):
        return self.start_plugin_future(chname, opname, None,
                                        alreadyOpenOk=alreadyOpenOk)

    def start_plugin_future(self, chname, opname, future,
                            alreadyOpenOk=True):
        try:
            p_info = self.get_plugin_info(opname)

        except KeyError:
            self.fv.show_error("No plugin information for plugin '%s'" % (
                opname))
            return

        if chname is not None:
            # local plugin
            plname = chname.upper() + ': ' + p_info.spec.get('tab', p_info.name)
            p_info.tabname = plname
        else:
            # global plugin
            plname = p_info.name
            p_info.tabname = p_info.spec.get('tab', plname)

        lname = p_info.name.lower()

        if lname in self.active:
            # <-- plugin is supposedly already active
            if alreadyOpenOk:
                self.set_focus(p_info.name)
                return

            raise PluginManagerError("Plugin %s is already active." % (
                plname))

        # Build GUI phase
        vbox = None
        try:
            if hasattr(p_info.obj, 'build_gui'):
                vbox = Widgets.VBox()

                in_ws = p_info.spec.get('workspace', None)
                if in_ws is None:
                    # to be deprecated
                    in_ws = p_info.spec.ws

                if in_ws.startswith('in:'):
                    # TODO: how to set this size appropriately
                    # Which plugins are actually using this attribute?
                    vbox.size = (400, 900)

                else:
                    # attach size of workspace to container so plugin
                    # can plan for how to configure itself
                    wd, ht = self.ds.get_ws_size(in_ws)
                    vbox.size = (wd, ht)

                if future:
                    p_info.obj.build_gui(vbox, future=future)
                else:
                    p_info.obj.build_gui(vbox)

        except Exception as e:
            errstr = "Plugin UI failed to initialize: %s" % (
                str(e))
            self.logger.error(errstr)
            try:
                (type, value, tb) = sys.exc_info()
                tb_str = "".join(traceback.format_tb(tb))
                self.logger.error("Traceback:\n%s" % (tb_str))

            except Exception as e:
                tb_str = "Traceback information unavailable."
                self.logger.error(tb_str)

            self.fv.show_error(errstr + '\n' + tb_str)
            #raise PluginManagerError(e)
            return

        # start phase
        try:
            if future:
                p_info.obj.start(future=future)
            else:
                p_info.obj.start()

        except Exception as e:
            errstr = "Plugin failed to start correctly: %s" % (
                str(e))
            self.logger.error(errstr)
            try:
                (type, value, tb) = sys.exc_info()
                tb_str = "".join(traceback.format_tb(tb))
                self.logger.error("Traceback:\n%s" % (tb_str))

            except Exception as e:
                tb_str = "Traceback information unavailable."
                self.logger.error(tb_str)

            self.fv.show_error(errstr + '\n' + tb_str)
            #raise PluginManagerError(e)
            return

        if vbox is not None:
            self.finish_gui(p_info, vbox)

            self.activate(p_info)
            self.set_focus(p_info.name)
        else:
            self.activate(p_info)
            # If this is a local plugin, raise the channel associated with the
            # plug in
            if p_info.chinfo is not None:
                itab = p_info.chinfo.name
                self.ds.raise_tab(itab)

    def stop_plugin(self, p_info):
        self.logger.debug("stopping plugin %s" % (str(p_info)))
        wasError = False
        e = None
        try:
            p_info.obj.stop()

        except Exception as e:
            wasError = True
            self.logger.error("Plugin failed to stop correctly: %s" % (
                str(e)))
            try:
                (type, value, tb) = sys.exc_info()
                tb_str = "".join(traceback.format_tb(tb))
                self.logger.error("Traceback:\n%s" % (tb_str))

            except Exception:
                self.logger.error("Traceback information unavailable.")

        if p_info.widget is not None:
            self.dispose_gui(p_info)
            self.ds.remove_tab(p_info.tabname)

        if wasError:
            raise PluginManagerError(e)

    def stop_all_plugins(self):
        for plugin_name in self.plugin:
            try:
                self.stop_plugin(plugin_name)
            except Exception as e:
                self.logger.error('Exception while calling stop for plugin %s: %s' % (plugin_name, str(e)))
        return True

    def plugin_build_error(self, box, text):
        textw = Widgets.TextArea(editable=False, wrap=True)
        textw.append_text(text)
        box.add_widget(textw, stretch=1)

    def finish_gui(self, p_info, vbox):
        # add container to workspace
        # TODO: how to figure out the appropriate size for top-levels?
        wd, ht = vbox.get_size()

        try:
            in_ws = p_info.spec.get('workspace', None)
            if in_ws is None:
                # to be deprecated
                in_ws = p_info.spec.ws

            if in_ws == 'in:toplevel':
                topw = vbox.get_app().make_window()
                topw.add_callback('close',
                                  lambda *args: self.deactivate(p_info.name))
                topw.resize(wd, ht)
                topw.set_widget(vbox)
                p_info.widget = topw
                p_info.is_toplevel = True
                topw.show()

            elif in_ws == 'in:dialog':
                dialog = Widgets.Dialog(title=p_info.name,
                                        flags=0,
                                        buttons=[],
                                        parent=self.fv.w.root)
                dialog.resize(wd, ht)
                box = dialog.get_content_area()
                box.add_widget(vbox, stretch=1)
                p_info.widget = dialog
                p_info.is_toplevel = True
                # TODO: need to add callback to remove from Desktop
                # dialog list?
                self.ds.show_dialog(dialog)

            else:
                bnch = self.ds.add_tab(in_ws, vbox, 2, p_info.tabname,
                                       p_info.tabname)
                bnch.plugin_info = p_info

                ws = self.ds.get_ws(in_ws)
                ws.add_callback('page-close', self.tab_closed_cb)

                ws_w = self.ds.get_nb(in_ws)
                ws_w.add_callback('page-switch', self.tab_switched_cb)
                p_info.widget = vbox
                p_info.is_toplevel = False

        except Exception as e:
            self.fv.show_error("Error finishing plugin UI for '%s': %s" % (
                p_info.name, str(e)))

    def tab_switched_cb(self, tab_w, widget):
        # A tab in a workspace in which we started a plugin has been
        # raised.  Check for this widget and focus the plugin
        title = widget.extdata.get('tab_title', None)
        # is this a local plugin tab?
        if title is None or ':' not in title:
            return

        chname, plname = title.split(':')
        plname = plname.strip()
        try:
            info = self.get_info(plname)
        except KeyError:
            # no
            return
        p_info = info.pInfo
        # important: make sure channel matches ours!
        if p_info.tabname == title:
            if self.is_active(p_info.name):
                if not self.has_focus(p_info.name):
                    self.set_focus(p_info.name)
                elif p_info.chinfo is not None:
                    # raise the channel associated with the plugin
                    chname = p_info.chinfo.name
                    # Alternative could just be to raise the channel tab
                    # rather than making all the global plugins switch
                    # over to the new channel
                    #self.ds.raise_tab(chname)
                    self.fv.change_channel(chname)

    def tab_closed_cb(self, ws, widget):
        bnch = self.ds._find_tab(widget)
        if bnch is not None:
            p_info = bnch.get('plugin_info', None)
            if p_info is not None:
                self.deactivate(p_info.name)

    def dispose_gui(self, p_info):
        self.logger.debug("disposing of gui")
        vbox = p_info.widget
        p_info.widget = None
        vbox.hide()
        vbox.delete()

    ########################################################
    ### NON-PEP8 PREDECESSORS: TO BE DEPRECATED

    loadPlugin = load_plugin
    reloadPlugin = reload_plugin
    getPluginInfo = get_plugin_info
    getPlugin = get_plugin
    getNames = get_names

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
Module wrapper for loading ASDF files.

.. note:: The API for this module is currently unstable
          and may change in a future release.

"""
import numpy as np

try:
    import asdf  # noqa
    from asdf.fits_embed import AsdfInFits
    have_asdf = True
except ImportError:
    have_asdf = False

from ginga import AstroImage
from ginga.util import iohelper, wcsmod
from ginga.util.io import io_base
# need this specific WCS for ASDF loads
from ginga.util.wcsmod import wcs_astropy_ape14   # noqa

__all__ = ['have_asdf', 'load_file', 'load_asdf', 'load_from_asdf',
           'ASDFFileHandler']


def load_from_asdf(asdf_obj, data_key='data', wcs_key='wcs', header_key='meta'):
    """
    Load from an ASDF object.  This method is primarily used internally
    to extract the correct parts of an ASDF file to reconstruct an image
    with WCS and metadata.

    Parameters
    ----------
    asdf_obj : obj
        ASDF or ASDF-in-FITS object.

    data_key, wcs_key, header_key : str
        Key values to specify where to find data, WCS, and header
        in ASDF.

    Returns
    -------
    data : ndarray or `None`
        Image data, if found.

    wcs : obj or `None`
        GWCS object or models, if found.

    ahdr : dict
        Header containing metadata.

    """
    asdf_keys = asdf_obj.keys()

    if wcs_key in asdf_keys:
        wcs = asdf_obj[wcs_key]
    elif wcs_key in asdf_obj[header_key]:
        wcs = asdf_obj[header_key][wcs_key]
    else:
        wcs = None

    if header_key in asdf_keys:
        ahdr = asdf_obj[header_key]
    else:
        ahdr = {}

    # TODO: What about non-image ASDF data, such as table?
    if data_key in asdf_keys:
        data = np.asarray(asdf_obj[data_key])
    # TODO: This hack is so earlier test files would still
    # load, but we need to make customization easier so
    # we can remove this hack.
    elif 'sci' in asdf_keys:
        data = np.asarray(asdf_obj['sci'])
    else:
        data = None

    return data, wcs, ahdr


def load_asdf(asdf_obj, idx=None, logger=None):
    """
    Load data from an open ASDF object.

    Parameters
    ----------
    asdf_obj : obj
        ASDF or ASDF-in-FITS object.

    idx : None
        Currently unused. Reserved for future use to identify specific
        data set of an ASDF object containing the data of interest.

    logger : python logging object
        Currently unused. Reserved for future use in logging

    Returns
    -------
    data : ndarray or `None`
        Image data, if found.

    wcs : obj or `None`
        GWCS object or models, if found.

    ahdr : dict
        Header containing metadata.

    """
    # TODO: support other types, like AstroTable, if ASDF can contain them
    data_obj = AstroImage.AstroImage(logger=logger)

    # TODO: idx may contain info about what part of the file to load
    #  e.g. should we pass as 'data_key' parameter?
    data, wcs, ahdr = load_from_asdf(asdf_obj)

    data_obj.setup_data(data, naxispath=None)

    wcsinfo = wcsmod.get_wcs_class('astropy_ape14')
    data_obj.wcs = wcsinfo.wrapper_class(logger=logger)
    data_obj.wcs.wcs = wcs

    if wcs is not None:
        data_obj.wcs.coordsys = wcs.output_frame.name

    return data_obj


def load_file(filepath, idx=None, logger=None, **kwargs):
    """
    Load an object from an ASDF file.
    See :func:`ginga.util.loader` for more info.

    """
    # see ginga.util.loader module
    opener = ASDFFileHandler(logger)
    data_obj = opener.load_file(filepath, idx=idx, **kwargs)

    return data_obj


class ASDFFileHandler(io_base.BaseIOHandler):

    name = 'asdf'

    def __init__(self, logger):
        super(ASDFFileHandler, self).__init__(logger)

        self._path = None

    def load_asdf(self, asdf_f, idx=None, **kwargs):

        data_obj = load_asdf(asdf_f, idx=idx, logger=self.logger)

        # set important metadata
        # TODO: is it possible to set the name without a filename?
        data_obj.set(idx=idx)
        data_obj.io = self

        return data_obj

    def load_asdf_hdu_in_fits(self, fits_f, hdu, **kwargs):
        """*** This is a special method that should only be called from
        WITHIN io_fits.py to open up ASDF-embedded-in-FITS ***
        """

        # Handle ASDF embedded in FITS (see load_hdu() in io_fits.py)
        # TODO: Populate EXTNAME, EXTVER, NAXISn in ASDF meta
        #   from HDU?
        # TODO: How to read from all the different ASDF layouts?
        # TODO: Cache the ASDF object?

        # TODO: hdu is ignored for now, but presumably this loader might
        # eventually want to check it
        with AsdfInFits.open(fits_f) as asdf_f:
            data_obj = self.load_asdf(asdf_f, logger=self.logger, **kwargs)

        # metadata will be hopefully be set back in io_fits
        return data_obj

    def load_file(self, filepath, idx=None, **kwargs):

        with asdf.open(filepath) as asdf_f:
            data_obj = load_asdf(asdf_f, idx=idx, logger=self.logger)

        # set important metadata
        name = iohelper.name_image_from_path(filepath, idx=None)
        data_obj.set(path=filepath, name=name, idx=idx,
                     image_loader=load_file)
        data_obj.io = self

        return data_obj

    def open_file(self, filepath, **kwargs):
        # TODO: this should open the ASDF file and make a full inventory
        # of what is available to load
        self._path = filepath
        return self

    def close(self):
        # TODO: this should close the ASDF file
        self._path = None

    def __len__(self):
        return 1

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def load_idx(self, idx, **kwargs):
        if self._path is None:
            raise ValueError("Please call open_file() first!")

        return self.load_file(self._path, idx=idx, **kwargs)

    def load_idx_cont(self, idx_spec, loader_cont_fn, **kwargs):

        data_obj = self.load_idx(None, **kwargs)

        # call continuation function
        loader_cont_fn(data_obj)



#
# images.py -- classes for images drawn on ginga canvases.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.canvas.CanvasObject import (CanvasObjectBase, _bool, _color,
                                       Point, MovePoint,
                                       register_canvas_types,
                                       colors_plus_none, coord_names)
from ginga.misc.ParamSet import Param
from ginga.misc import Bunch
from ginga import trcalc

from .mixins import OnePointMixin


class Image(OnePointMixin, CanvasObjectBase):
    """Draws an image on a ImageViewCanvas.
    Parameters are:
    x, y: 0-based coordinates of one corner in the data space
    image: the image, which must be an RGBImage object
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of corner of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of corner of object"),
            ## Param(name='image', type=?, argpos=2,
            ##       description="Image to be displayed on canvas"),
            Param(name='scale_x', type=float, default=1.0,
                  description="Scaling factor for X dimension of object"),
            Param(name='scale_y', type=float, default=1.0,
                  description="Scaling factor for Y dimension of object"),
            Param(name='interpolation', type=str, default='basic',
                  description="Interpolation method for scaling pixels"),
            Param(name='linewidth', type=int, default=0,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default: solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='lightgreen',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            Param(name='flipy', type=_bool,
                  default=False, valid=[False, True],
                  description="Flip image in Y direction"),
            Param(name='optimize', type=_bool,
                  default=True, valid=[False, True],
                  description="Optimize rendering for this object"),
        ]

    def __init__(self, x, y, image, alpha=1.0, scale_x=1.0, scale_y=1.0,
                 interpolation='basic',
                 linewidth=0, linestyle='solid', color='lightgreen',
                 showcap=False, flipy=False, optimize=True,
                 **kwdargs):
        self.kind = 'image'
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, image=image, alpha=alpha,
                                  scale_x=scale_x, scale_y=scale_y,
                                  interpolation=interpolation,
                                  linewidth=linewidth, linestyle=linestyle,
                                  color=color, showcap=showcap,
                                  flipy=flipy, optimize=optimize,
                                  **kwdargs)
        OnePointMixin.__init__(self)

        # The cache holds intermediate step results by viewer.
        # Depending on value of `whence` they may not need to be recomputed.
        self._cache = {}
        self._zorder = 0
        # images are not editable by default
        self.editable = False

        self.enable_callback('image-set')

    def get_zorder(self):
        return self._zorder

    def set_zorder(self, zorder):
        self._zorder = zorder
        for viewer in self._cache:
            viewer.reorder_layers()
            viewer.redraw(whence=2)

    def in_cache(self, viewer):
        return viewer in self._cache

    def get_cache(self, viewer):
        if viewer in self._cache:
            cache = self._cache[viewer]
        else:
            cache = self._reset_cache(Bunch.Bunch())
            self._cache[viewer] = cache
        return cache

    def invalidate_cache(self, viewer):
        cache = self.get_cache(viewer)
        self._reset_cache(cache)
        return cache

    def draw(self, viewer):
        """General draw method for RGB image types.
        Note that actual insertion of the image into the output is
        handled in `draw_image()`
        """
        cache = self.get_cache(viewer)

        if not cache.drawn:
            cache.drawn = True
            viewer.redraw(whence=2)

        cpoints = self.get_cpoints(viewer)
        cr = viewer.renderer.setup_cr(self)

        # draw optional border
        if self.linewidth > 0:
            cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)

    def draw_image(self, viewer, dstarr, whence=0.0):
        if self.image is None:
            return

        cache = self.get_cache(viewer)

        dst_order = viewer.get_rgb_order()
        image_order = self.image.get_order()

        if (whence <= 0.0) or (cache.cutout is None) or (not self.optimize):
            # get extent of our data coverage in the window
            pts = np.asarray(viewer.get_pan_rect()).T
            xmin = int(np.min(pts[0]))
            ymin = int(np.min(pts[1]))
            xmax = int(np.ceil(np.max(pts[0])))
            ymax = int(np.ceil(np.max(pts[1])))

            # get destination location in data_coords
            dst_x, dst_y = self.crdmap.to_data((self.x, self.y))

            a1, b1, a2, b2 = 0, 0, self.image.width - 1, self.image.height - 1

            # calculate the cutout that we can make and scale to merge
            # onto the final image--by only cutting out what is necessary
            # this speeds scaling greatly at zoomed in sizes
            ((dst_x, dst_y), (a1, b1), (a2, b2)) = \
                trcalc.calc_image_merge_clip((xmin, ymin), (xmax, ymax),
                                             (dst_x, dst_y),
                                             (a1, b1), (a2, b2))

            # is image completely off the screen?
            if (a2 - a1 <= 0) or (b2 - b1 <= 0):
                # no overlay needed
                return

            # cutout and scale the piece appropriately by the viewer scale
            scale_x, scale_y = viewer.get_scale_xy()
            # scale additionally by our scale
            _scale_x, _scale_y = scale_x * self.scale_x, scale_y * self.scale_y

            res = self.image.get_scaled_cutout2((a1, b1), (a2, b2),
                                                (_scale_x, _scale_y),
                                                method=self.interpolation)

            # don't ask for an alpha channel from overlaid image if it
            # doesn't have one
            ## if ('A' in dst_order) and not ('A' in image_order):
            ##     dst_order = dst_order.replace('A', '')

            ## if dst_order != image_order:
            ##     # reorder result to match desired rgb_order by backend
            ##     cache.cutout = trcalc.reorder_image(dst_order, res.data,
            ##                                          image_order)
            ## else:
            ##     cache.cutout = res.data
            data = res.data
            if self.flipy:
                data = np.flipud(data)
            cache.cutout = data

            # calculate our offset from the pan position
            pan_x, pan_y = viewer.get_pan()
            pan_off = viewer.data_off
            pan_x, pan_y = pan_x + pan_off, pan_y + pan_off
            off_x, off_y = dst_x - pan_x, dst_y - pan_y
            # scale offset
            off_x *= scale_x
            off_y *= scale_y

            # dst position in the pre-transformed array should be calculated
            # from the center of the array plus offsets
            ht, wd, dp = dstarr.shape
            cvs_x = int(np.round(wd / 2.0 + off_x))
            cvs_y = int(np.round(ht / 2.0 + off_y))
            cache.cvs_pos = (cvs_x, cvs_y)

        # composite the image into the destination array at the
        # calculated position
        trcalc.overlay_image(dstarr, cache.cvs_pos, cache.cutout,
                             dst_order=dst_order, src_order=image_order,
                             alpha=self.alpha, fill=True, flipy=False)

    def _reset_cache(self, cache):
        cache.setvals(cutout=None, drawn=False, cvs_pos=(0, 0))
        return cache

    def reset_optimize(self):
        for cache in self._cache.values():
            self._reset_cache(cache)

    def get_image(self):
        return self.image

    def set_image(self, image):
        self.image = image
        self.reset_optimize()

        self.make_callback('image-set', image)

    def get_scaled_wdht(self):
        width = int(self.image.width * self.scale_x)
        height = int(self.image.height * self.scale_y)
        return (width, height)

    def get_coords(self):
        x1, y1 = self.crdmap.to_data((self.x, self.y))
        wd, ht = self.get_scaled_wdht()
        x2, y2 = x1 + wd, y1 + ht
        return (x1, y1, x2, y2)

    def get_llur(self):
        return self.get_coords()

    def get_center_pt(self):
        wd, ht = self.get_scaled_wdht()
        x1, y1, x2, y2 = self.get_coords()
        return ((x1 + x2) / 2.0, (y1 + y2) / 2.0)

    def get_points(self):
        x1, y1, x2, y2 = self.get_coords()
        return [(x1, y1), (x2, y1), (x2, y2), (x1, y2)]

    def contains_pt(self, pt):
        data_x, data_y = pt[:2]
        x1, y1, x2, y2 = self.get_coords()
        if ((x1 <= data_x < x2) and (y1 <= data_y < y2)):
            return True
        return False

    def rotate(self, theta, xoff=0, yoff=0):
        raise ValueError("Images cannot be rotated")

    def setup_edit(self, detail):
        detail.center_pos = self.get_center_pt()
        detail.scale_x = self.scale_x
        detail.scale_y = self.scale_y

    def set_edit_point(self, i, pt, detail):
        if i == 0:
            self.move_to_pt(pt)
        elif i == 1:
            scale_x, scale_y = self.calc_dual_scale_from_pt(pt, detail)
            self.scale_x = detail.scale_x * scale_x
        elif i == 2:
            scale_x, scale_y = self.calc_dual_scale_from_pt(pt, detail)
            self.scale_y = detail.scale_y * scale_y
        elif i == 3:
            scale_x, scale_y = self.calc_dual_scale_from_pt(pt, detail)
            self.scale_x = detail.scale_x * scale_x
            self.scale_y = detail.scale_y * scale_y
        else:
            raise ValueError("No point corresponding to index %d" % (i))

        self.reset_optimize()

    def get_edit_points(self, viewer):
        x1, y1, x2, y2 = self.get_coords()
        return [MovePoint(*self.get_center_pt()),    # location
                Point(x2, (y1 + y2) / 2.),   # width scale
                Point((x1 + x2) / 2., y2),   # height scale
                Point(x2, y2),               # both scale
                ]

    def scale_by_factors(self, factors):
        scale_x, scale_y = factors[:2]
        self.scale_x *= scale_x
        self.scale_y *= scale_y
        self.reset_optimize()

    def set_scale(self, scale_x, scale_y):
        self.scale_x = scale_x
        self.scale_y = scale_y
        self.reset_optimize()

    def set_origin(self, x, y):
        self.x, self.y = x, y
        self.reset_optimize()


class NormImage(Image):
    """Draws an image on a ImageViewCanvas.

    Parameters are:
    x, y: 0-based coordinates of one corner in the data space
    image: the image, which must be an RGBImage object
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of corner of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of corner of object"),
            ## Param(name='image', type=?, argpos=2,
            ##       description="Image to be displayed on canvas"),
            Param(name='scale_x', type=float, default=1.0,
                  description="Scaling factor for X dimension of object"),
            Param(name='scale_y', type=float, default=1.0,
                  description="Scaling factor for Y dimension of object"),
            Param(name='interpolation', type=str, default='basic',
                  description="Interpolation method for scaling pixels"),
            Param(name='linewidth', type=int, default=0,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default: solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='lightgreen',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            ## Param(name='flipy', type=_bool,
            ##       default=False, valid=[False, True],
            ##       description="Flip image in Y direction"),
            Param(name='optimize', type=_bool,
                  default=True, valid=[False, True],
                  description="Optimize rendering for this object"),
            ## Param(name='rgbmap', type=?,
            ##       description="RGB mapper for the image"),
            ## Param(name='autocuts', type=?,
            ##       description="Cuts manager for the image"),
        ]

    def __init__(self, x, y, image, alpha=1.0, scale_x=1.0, scale_y=1.0,
                 interpolation='basic', linewidth=0, linestyle='solid',
                 color='lightgreen', showcap=False,
                 optimize=True, rgbmap=None, autocuts=None, **kwdargs):
        self.kind = 'normimage'
        super(NormImage, self).__init__(x, y, image=image, alpha=alpha,
                                        scale_x=scale_x, scale_y=scale_y,
                                        interpolation=interpolation,
                                        linewidth=linewidth, linestyle=linestyle,
                                        color=color,
                                        showcap=showcap, optimize=optimize,
                                        **kwdargs)
        self.rgbmap = rgbmap
        self.autocuts = autocuts

    def draw_image(self, viewer, dstarr, whence=0.0):
        if self.image is None:
            return

        cache = self.get_cache(viewer)

        if (whence <= 0.0) or (cache.cutout is None) or (not self.optimize):
            # get extent of our data coverage in the window
            pts = np.asarray(viewer.get_pan_rect()).T
            xmin = int(np.min(pts[0]))
            ymin = int(np.min(pts[1]))
            xmax = int(np.ceil(np.max(pts[0])))
            ymax = int(np.ceil(np.max(pts[1])))

            # destination location in data_coords
            dst_x, dst_y = self.crdmap.to_data((self.x, self.y))

            a1, b1, a2, b2 = 0, 0, self.image.width - 1, self.image.height - 1

            # calculate the cutout that we can make and scale to merge
            # onto the final image--by only cutting out what is necessary
            # this speeds scaling greatly at zoomed in sizes
            ((dst_x, dst_y), (a1, b1), (a2, b2)) = \
                trcalc.calc_image_merge_clip((xmin, ymin), (xmax, ymax),
                                             (dst_x, dst_y),
                                             (a1, b1), (a2, b2))

            # is image completely off the screen?
            if (a2 - a1 <= 0) or (b2 - b1 <= 0):
                # no overlay needed
                return

            # cutout and scale the piece appropriately by viewer scale
            scale_x, scale_y = viewer.get_scale_xy()
            # scale additionally by our scale
            _scale_x, _scale_y = scale_x * self.scale_x, scale_y * self.scale_y

            res = self.image.get_scaled_cutout2((a1, b1), (a2, b2),
                                                (_scale_x, _scale_y),
                                                method=self.interpolation)
            cache.cutout = res.data

            # calculate our offset from the pan position
            pan_x, pan_y = viewer.get_pan()
            pan_off = viewer.data_off
            pan_x, pan_y = pan_x + pan_off, pan_y + pan_off
            off_x, off_y = dst_x - pan_x, dst_y - pan_y
            # scale offset
            off_x *= scale_x
            off_y *= scale_y

            # dst position in the pre-transformed array should be calculated
            # from the center of the array plus offsets
            ht, wd, dp = dstarr.shape
            cvs_x = int(np.round(wd / 2.0 + off_x))
            cvs_y = int(np.round(ht / 2.0 + off_y))
            cache.cvs_pos = (cvs_x, cvs_y)

        if self.rgbmap is not None:
            rgbmap = self.rgbmap
        else:
            rgbmap = viewer.get_rgbmap()

        if (whence <= 1.0) or (cache.prergb is None) or (not self.optimize):
            # apply visual changes prior to color mapping (cut levels, etc)
            vmax = rgbmap.get_hash_size() - 1
            newdata = self.apply_visuals(viewer, cache.cutout, 0, vmax)

            # result becomes an index array fed to the RGB mapper
            if not np.issubdtype(newdata.dtype, np.dtype('uint')):
                newdata = newdata.astype(np.uint)
            idx = newdata

            self.logger.debug("shape of index is %s" % (str(idx.shape)))
            cache.prergb = idx

        dst_order = viewer.get_rgb_order()
        image_order = self.image.get_order()
        get_order = dst_order
        # note: is this still needed?  I think overlay_image will handle
        # a mismatch of alpha channel now
        if ('A' in dst_order) and not ('A' in image_order):
            get_order = dst_order.replace('A', '')

        if (whence <= 2.5) or (cache.rgbarr is None) or (not self.optimize):
            # get RGB mapped array
            rgbobj = rgbmap.get_rgbarray(cache.prergb, order=dst_order,
                                         image_order=image_order)
            cache.rgbarr = rgbobj.get_array(get_order)

        # composite the image into the destination array at the
        # calculated position
        trcalc.overlay_image(dstarr, cache.cvs_pos, cache.rgbarr,
                             dst_order=dst_order, src_order=get_order,
                             alpha=self.alpha, fill=True, flipy=False)

    def apply_visuals(self, viewer, data, vmin, vmax):
        if self.autocuts is not None:
            autocuts = self.autocuts
        else:
            autocuts = viewer.autocuts

        # Apply cut levels
        loval, hival = viewer.t_['cuts']
        newdata = autocuts.cut_levels(data, loval, hival,
                                      vmin=vmin, vmax=vmax)
        return newdata

    def _reset_cache(self, cache):
        cache.setvals(cutout=None, prergb=None, rgbarr=None,
                      drawn=False, cvs_pos=(0, 0))
        return cache

    def set_image(self, image):
        self.image = image
        self.reset_optimize()

        self.make_callback('image-set', image)

    def scale_by(self, scale_x, scale_y):
        self.scale_x *= scale_x
        self.scale_y *= scale_y
        self.reset_optimize()


# register our types
register_canvas_types(dict(image=Image, normimage=NormImage))

#END



#
# ImageViewCanvasMock.py -- A Ginga image widget with canvas drawing in mock
#                             widget set
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.mockw import ImageViewMock
from ginga.canvas.mixins import DrawingMixin, CanvasMixin, CompoundMixin


class ImageViewCanvasError(ImageViewMock.ImageViewMockError):
    pass


class ImageViewCanvas(ImageViewMock.ImageViewZoom,
                      DrawingMixin, CanvasMixin, CompoundMixin):

    def __init__(self, logger=None, settings=None,
                 rgbmap=None, bindmap=None, bindings=None):
        ImageViewMock.ImageViewZoom.__init__(self, logger=logger,
                                             settings=settings,
                                             rgbmap=rgbmap,
                                             bindmap=bindmap,
                                             bindings=bindings)
        CompoundMixin.__init__(self)
        CanvasMixin.__init__(self)
        DrawingMixin.__init__(self)

        # we are both a viewer and a canvas
        self.set_canvas(self, private_canvas=self)

    # METHODS THAT WERE IN IPG

    def add_canvas(self, tag=None):
        # add a canvas to the view
        DrawingCanvas = self.getDrawClass('drawingcanvas')
        canvas = DrawingCanvas()
        # enable drawing on the canvas
        canvas.enable_draw(True)
        canvas.ui_set_active(True)
        canvas.set_surface(self)
        # add the canvas to the view.
        self.add(canvas, tag=tag)
        return canvas

    def show(self):
        from IPython.display import Image
        return Image(data=bytes(self.get_rgb_image_as_bytes(format='png')),
                     format='png', embed=True)

#END



#
# geometry_helper.py -- help module for Ginga OpenGL camera
#
# Credit:
#   Modified from code written by M. McGuffin
#   http://profs.etsmtl.ca/mmcguffin/code/python/example-3D_Python-Qt-OpenGL/
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np


class Point3D(object):

    def __init__(self, x=0, y=0, z=0):
        self.coord = np.array([x, y, z], dtype=np.float32)

    @property
    def x(self):
        return self.coord[0]

    @property
    def y(self):
        return self.coord[1]

    @property
    def z(self):
        return self.coord[2]

    def __repr__(self):
        return "Point3D(%f, %f, %f)" % tuple(self.coord)

    def __str__(self):
        return "P(%f, %f, %f)" % tuple(self.coord)

    def get(self):
        return self.coord

    def copy(self):
        return Point3D(self.x, self.y, self.z)

    def as_Vector3D(self):
        return Vector3D(self.x, self.y, self.z)

    def distance(self, other):
        return (other - self).length()

    def average(self, other):
        return Point3D((self.x + other.x) * 0.5,
                       (self.y + other.y) * 0.5,
                       (self.z + other.z) * 0.5)

    def __add__(self, other):
        return Point3D(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        if isinstance(other, Vector3D):
            return Point3D(self.x - other.x, self.y - other.y, self.z - other.z)

        return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __ne__(self, other):
        return not (self == other)


class Vector3D(object):

    def __init__(self, x=0, y=0, z=0):
        self.coord = np.array([x, y, z], dtype=np.float32)

    @property
    def x(self):
        return self.coord[0]

    @property
    def y(self):
        return self.coord[1]

    @property
    def z(self):
        return self.coord[2]

    def __repr__(self):
        return "Vector3D(%f, %f, %f)" % tuple(self.coord)

    def __str__(self):
        return "V(%f, %f, %f)" % tuple(self.coord)

    def get(self):
        return self.coord

    def copy(self):
        return Vector3D(self.x, self.y, self.z)

    def as_Point3D(self):
        return Point3D(self.x, self.y, self.z)

    def length_squared(self):
        return self.x * self.x + self.y * self.y + self.z * self.z

    def length(self):
        return np.sqrt(self.length_squared())

    def normalized(self):
        l = self.length()
        if l > 0:
            return Vector3D(self.x / l, self.y / l, self.z / l)
        return self.copy()

    def __neg__(self):
        return Vector3D(-self.x, -self.y, -self.z)

    def __add__(self, other):
        if isinstance(other, Point3D):
            return Point3D(self.x + other.x, self.y + other.y, self.z + other.z)

        return Vector3D(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other):
        if isinstance(other, Vector3D):
            # dot product
            return self.x * other.x + self.y * other.y + self.z * other.z

        # scalar product
        return Vector3D(self.x * other, self.y * other, self.z * other)

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        return Vector3D(self.x / other, self.y / other, self.z / other)

    def __xor__(self, other):   # cross product
        return Vector3D(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __ne__(self, other):
        return not (self == other)


class Matrix4x4(object):

    def __init__(self):
        self.set_to_identity()

    def __str__(self):
        return str(self.m)

    @property
    def f(self):
        return self.m.flat

    def get(self):
        return self.m

    def copy(self):
        M = Matrix4x4()
        M.m = self.m.copy()
        return M

    def set_to_identity(self):
        self.m = np.eye(4)

    @staticmethod
    def translation(vector3D):
        M = Matrix4x4()
        M.f[12] = vector3D.x
        M.f[13] = vector3D.y
        M.f[14] = vector3D.z
        return M

    @staticmethod
    def rotation_around_origin(angle_rad, axis_vector):
        # Note: assumes axis_vector is normalized
        c = np.cos(angle_rad)
        s = np.sin(angle_rad)
        one_minus_c = 1 - c
        M = Matrix4x4()
        M.f[0] = c + one_minus_c * axis_vector.x * axis_vector.x
        M.f[5] = c + one_minus_c * axis_vector.y * axis_vector.y
        M.f[10] = c + one_minus_c * axis_vector.z * axis_vector.z
        M.f[1] = M.f[4] = one_minus_c * axis_vector.x * axis_vector.y
        M.f[2] = M.f[8] = one_minus_c * axis_vector.x * axis_vector.z
        M.f[6] = M.f[9] = one_minus_c * axis_vector.y * axis_vector.z
        xs = axis_vector.x * s
        ys = axis_vector.y * s
        zs = axis_vector.z * s
        M.f[1] += zs
        M.f[4] -= zs
        M.f[2] -= ys
        M.f[8] += ys
        M.f[6] += xs
        M.f[9] -= xs

        M.f[12] = 0.0
        M.f[13] = 0.0
        M.f[14] = 0.0
        M.f[3] = 0.0
        M.f[7] = 0.0
        M.f[11] = 0.0
        M.f[15] = 1.0
        return M

    @staticmethod
    def rotation(angle_rad, axis_vector, origin_point):
        v = origin_point.as_Vector3D()
        return Matrix4x4.translation(v) * Matrix4x4.rotation_around_origin(angle_rad, axis_vector) * Matrix4x4.translation(- v)

    @staticmethod
    def uniform_scale_around_origin(scale_factor):
        M = Matrix4x4()
        M.m *= scale_factor
        M.f[15] = 1.0
        return M

    @staticmethod
    def uniform_scale(scale_factor, origin_point):
        v = origin_point.as_Vector3D()
        return Matrix4x4.translation(v) * Matrix4x4.uniform_scale_around_origin(scale_factor) * Matrix4x4.translation(- v)

    @staticmethod
    def look_at(eye_point, target_point, up_vector, is_inverted):
        # step one: generate a rotation matrix

        z = (eye_point - target_point).normalized()
        y = up_vector
        x = y ^ z   # cross product
        y = z ^ x   # cross product

        # Cross product gives area of parallelogram, which is < 1 for
        # non-perpendicular unit-length vectors; so normalize x and y.
        x = x.normalized()
        y = y.normalized()

        M = Matrix4x4()

        if is_inverted:
            # the rotation matrix
            M.f[0] = x.x
            M.f[4] = y.x
            M.f[8] = z.x
            M.f[12] = 0.0
            M.f[1] = x.y
            M.f[5] = y.y
            M.f[9] = z.y
            M.f[13] = 0.0
            M.f[2] = x.z
            M.f[6] = y.z
            M.f[10] = z.z
            M.f[14] = 0.0
            M.f[3] = 0.0
            M.f[7] = 0.0
            M.f[11] = 0.0
            M.f[15] = 1.0

            # step two: premultiply by a translation matrix
            return Matrix4x4.translation(eye_point.as_Vector3D()) * M

        else:
            # the rotation matrix
            M.f[0] = x.x
            M.f[4] = x.y
            M.f[8] = x.z
            M.f[12] = 0.0
            M.f[1] = y.x
            M.f[5] = y.y
            M.f[9] = y.z
            M.f[13] = 0.0
            M.f[2] = z.x
            M.f[6] = z.y
            M.f[10] = z.z
            M.f[14] = 0.0
            M.f[3] = 0.0
            M.f[7] = 0.0
            M.f[11] = 0.0
            M.f[15] = 1.0

            # step two: postmultiply by a translation matrix
            return M * Matrix4x4.translation(- eye_point.as_Vector3D())

    def __mul__(self, b):
        a = self
        if isinstance(b, Matrix4x4):
            M = Matrix4x4()
            M.f[0] = (a.f[0] * b.f[0] + a.f[4] * b.f[1] +
                      a.f[8] * b.f[2] + a.f[12] * b.f[3])
            M.f[1] = (a.f[1] * b.f[0] + a.f[5] * b.f[1] +
                      a.f[9] * b.f[2] + a.f[13] * b.f[3])
            M.f[2] = (a.f[2] * b.f[0] + a.f[6] * b.f[1] +
                      a.f[10] * b.f[2] + a.f[14] * b.f[3])
            M.f[3] = (a.f[3] * b.f[0] + a.f[7] * b.f[1] +
                      a.f[11] * b.f[2] + a.f[15] * b.f[3])

            M.f[4] = (a.f[0] * b.f[4] + a.f[4] * b.f[5] +
                      a.f[8] * b.f[6] + a.f[12] * b.f[7])
            M.f[5] = (a.f[1] * b.f[4] + a.f[5] * b.f[5] +
                      a.f[9] * b.f[6] + a.f[13] * b.f[7])
            M.f[6] = (a.f[2] * b.f[4] + a.f[6] * b.f[5] +
                      a.f[10] * b.f[6] + a.f[14] * b.f[7])
            M.f[7] = (a.f[3] * b.f[4] + a.f[7] * b.f[5] +
                      a.f[11] * b.f[6] + a.f[15] * b.f[7])

            M.f[8] = (a.f[0] * b.f[8] + a.f[4] * b.f[9] +
                      a.f[8] * b.f[10] + a.f[12] * b.f[11])
            M.f[9] = (a.f[1] * b.f[8] + a.f[5] * b.f[9] +
                      a.f[9] * b.f[10] + a.f[13] * b.f[11])
            M.f[10] = (a.f[2] * b.f[8] + a.f[6] * b.f[9] +
                       a.f[10] * b.f[10] + a.f[14] * b.f[11])
            M.f[11] = (a.f[3] * b.f[8] + a.f[7] * b.f[9] +
                       a.f[11] * b.f[10] + a.f[15] * b.f[11])

            M.f[12] = (a.f[0] * b.f[12] + a.f[4] * b.f[13] +
                       a.f[8] * b.f[14] + a.f[12] * b.f[15])
            M.f[13] = (a.f[1] * b.f[12] + a.f[5] * b.f[13] +
                       a.f[9] * b.f[14] + a.f[13] * b.f[15])
            M.f[14] = (a.f[2] * b.f[12] + a.f[6] * b.f[13] +
                       a.f[10] * b.f[14] + a.f[14] * b.f[15])
            M.f[15] = (a.f[3] * b.f[12] + a.f[7] * b.f[13] +
                       a.f[11] * b.f[14] + a.f[15] * b.f[15])
            return M

        elif isinstance(b, Vector3D):
            # We treat the vector as if its (homogeneous) 4th component were zero.
            return Vector3D(
                a.f[0] * b.x + a.f[4] * b.y + a.f[8] * b.z,  # + a.f[12]*b.w(),
                a.f[1] * b.x + a.f[5] * b.y + a.f[9] * b.z,  # + a.f[13]*b.w(),
                a.f[2] * b.x + a.f[6] * b.y + a.f[10] * b.z  # + a.f[14]*b.w(),
                # a.f[3] * b.x + a.f[7] * b.y + a.f[11] * b.z + a.f[15]*b.w()
            )

        elif isinstance(b, Point3D):
            # We treat the point as if its (homogeneous) 4th component were one.
            return Point3D(
                a.f[0] * b.x + a.f[4] * b.y + a.f[8] * b.z + a.f[12],
                a.f[1] * b.x + a.f[5] * b.y + a.f[9] * b.z + a.f[13],
                a.f[2] * b.x + a.f[6] * b.y + a.f[10] * b.z + a.f[14]
            )

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
Web browser plugin for Ginga.

**Plugin Type: Global**

``WBrowser`` is a global plugin. Only one instance can be opened.

**Usage**

This global plugin is used to browse help pages for Ginga.

When a "Help" button is pressed from a plugin (e.g., ``Pick``),
Ginga will attempt to download an existing documentation build
from *ReadTheDocs* for the matching version. If successful,
plugin documentation from that download is displayed.
If not successful or deliberately disabled in "plugin_WBrowser.cfg",
Ginga will render the plugin's docstring locally.

"""

import os

# GINGA
from ginga.GingaPlugin import GlobalPlugin
from ginga.gw import Widgets

__all__ = ['WBrowser']

WAIT_HTML = """<html><title>Downloading documentation</title>
<body>
<p>Downloading and unpacking Ginga documentation from ReadTheDocs.
This may take several seconds (or longer, depending on your connection).</p>
<p>Please wait...</p>
</body>
</html>
"""


class WBrowser(GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(WBrowser, self).__init__(fv)

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_WBrowser')
        self.settings.add_defaults(offline_doc_only=False)
        self.settings.load(onError='silent')

    def build_gui(self, container):
        if not Widgets.has_webkit:
            self.browser = Widgets.Label(
                "Please install the python-webkit package to enable "
                "this plugin")
        else:
            self.browser = Widgets.WebView()

        sw = Widgets.ScrollArea()
        sw.set_widget(self.browser)

        container.add_widget(sw, stretch=1)
        sw.show()

        self.entry = Widgets.TextEntrySet()
        container.add_widget(self.entry, stretch=0)
        self.entry.add_callback('activated', lambda w: self.browse_cb())

        tbar = Widgets.Toolbar(orientation='horizontal')
        for tt, cb, ico in (
                ('Go back', lambda w: self.back_cb(), 'prev_48.png'),
                ('Go forward', lambda w: self.forward_cb(), 'next_48.png'),
                ('Reload page', lambda w: self.reload_cb(), 'rotate_48.png'),
                ('Stop loading', lambda w: self.stop_cb(), 'stop_48.png'),
                ('Go to top of documentation', lambda w: self.show_help(),
                 'fits.png')):
            btn = tbar.add_action(
                None, iconpath=os.path.join(self.fv.iconpath, ico))
            btn.add_callback('activated', cb)
            btn.set_tooltip(tt)
        container.add_widget(tbar, stretch=0)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(3)
        btn = Widgets.Button('Close')
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button('Help')
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        container.add_widget(btns, stretch=0)

        self.gui_up = True

    def _download_doc(self, plugin=None, no_url_callback=None):
        from ginga.doc.download_doc import get_doc
        self.fv.assert_nongui_thread()

        self.fv.gui_do(self._load_doc, WAIT_HTML, url_is_content=True)

        def _dl_indicator(count, blksize, totalsize):
            pct = (count * blksize) / totalsize
            msg = 'Downloading: {:.1%} complete'.format(pct)
            self.fv.gui_do(self.entry.set_text, msg)

        # This can block as long as it takes without blocking the UI.
        if self.settings.get('offline_doc_only', False):
            url = None  # DEBUG: Use this to force offline mode.
        else:
            url = get_doc(logger=self.logger, plugin=plugin,
                          reporthook=_dl_indicator)

        self.fv.gui_do(self._load_doc, url, no_url_callback=no_url_callback)

    def _load_doc(self, url, no_url_callback=None, url_is_content=False):
        self.fv.assert_gui_thread()
        if url is None:
            self.entry.set_text('')
            if no_url_callback is None:
                self.fv.show_error("Couldn't load web page")
            else:
                no_url_callback()
        else:
            self.browse(url, url_is_content=url_is_content)

    def show_help(self, plugin=None, no_url_callback=None):
        """See `~ginga.GingaPlugin` for usage of optional keywords."""
        if not Widgets.has_webkit:
            return

        self.fv.nongui_do(self._download_doc, plugin=plugin,
                          no_url_callback=no_url_callback)

    def browse(self, url, url_is_content=False):
        if not Widgets.has_webkit:
            return

        try:
            if url_is_content:  # This was load_html()
                self.browser.load_html_string(url)
                self.entry.set_text('')
            else:
                self.logger.debug("Browsing '{}'".format(url))
                self.browser.load_url(url)
                self.entry.set_text(url)
        except Exception as e:
            self.fv.show_error("Couldn't load web page: {}".format(str(e)))
        else:
            self.browser.show()

    def browse_cb(self):
        url = str(self.entry.get_text()).strip()
        if len(url) > 0:
            self.browse(url)

    def back_cb(self):
        if not Widgets.has_webkit:
            return
        self.browser.go_back()

    def forward_cb(self):
        if not Widgets.has_webkit:
            return
        self.browser.go_forward()

    def reload_cb(self):
        if not Widgets.has_webkit:
            return
        self.browser.reload_page()

    def stop_cb(self):
        if not Widgets.has_webkit:
            return
        self.browser.stop_loading()

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'wbrowser'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_WBrowser', package='ginga')

# END



#
# Camera.py -- Ginga 3D viewer OpenGL camera
#
# Credit:
#   Modified from code written by M. McGuffin
#   http://profs.etsmtl.ca/mmcguffin/code/python/example-3D_Python-Qt-OpenGL/
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np
from OpenGL import GL as gl

from .geometry_helper import Point3D, Vector3D, Matrix4x4


class Camera(object):

    def __init__(self):
        self.fov_deg = 30.0
        # Orbiting speed in degrees per radius of viewport
        self.orbit_speed = 300.0

        # These are in world-space units.
        self.near_plane = 1.0
        self.far_plane = 10000.0

        # During dollying (i.e. when the camera is translating into
        # the scene), if the camera gets too close to the target
        # point, we push the target point away.
        # The threshold distance at which such "pushing" of the
        # target point begins is this fraction of near_plane.
        # To prevent the target point from ever being clipped,
        # this fraction should be chosen to be greater than 1.0.
        self.push_threshold = 1.3

        # Viewport dimensions and radius (in pixels)
        # We give these some initial values just as a safeguard
        # against division by zero when computing their ratio.
        self.vport_wd_px = 10
        self.vport_ht_px = 10
        self.vport_radius_px = 5

        self.scene_radius = 10

        # point of view, or center of camera; the ego-center; the eye-point
        self.position = Point3D()

        # point of interest; what the camera is looking at; the exo-center
        self.target = Point3D()

        # This is the up vector for the (local) camera space
        self.up = Vector3D()

        # This is the up vector for the (global) world space;
        # it is perpendicular to the horizontal (x,z)-plane
        self.ground = Vector3D(0, 1, 0)

        # this is the default eye position to reset the camera
        self.home_position = Point3D(0, 0, 1)

        # this is the default target position to reset the camera
        self.tgt_position = Point3D(0, 0, 0)

        self.reset()

    def reset(self):
        #self.position = Point3D(0, 0, -distance_from_target)
        self.position = self.home_position.copy()
        self.target = self.tgt_position.copy()
        self.up = self.ground.copy()

    def set_viewport_dimensions(self, wd_px, ht_px):
        self.vport_wd_px = wd_px
        self.vport_ht_px = ht_px
        self.vport_radius_px = 0.5 * wd_px if (wd_px < ht_px) else 0.5 * ht_px

    def get_viewport_dimensions(self):
        return (self.vport_wd_px, self.vport_ht_px)

    def set_scene_radius(self, radius):
        self.scene_radius = radius

        # recalculate home position
        x, y, z = self.home_position.get()

        tangent = np.tan(self.fov_deg / 2.0 / 180.0 * np.pi)
        distance_from_target = self.scene_radius / tangent
        self.home_position = Point3D(x, y, distance_from_target)

    def set_camera_home_position(self, pt):
        self.home_position = Point3D(*pt)

    def set_target_home_position(self, pt):
        self.tgt_position = Point3D(*pt)

    def save_positions(self):
        self.home_position = self.position.copy()
        self.tgt_position = self.target.copy()

    def set_gl_transform(self):
        """This side effects the OpenGL context to set the view to match
        the camera.
        """
        tangent = np.tan(self.fov_deg / 2.0 / 180.0 * np.pi)
        vport_radius = self.near_plane * tangent
        # calculate aspect of the viewport
        if self.vport_wd_px < self.vport_ht_px:
            vport_wd = 2.0 * vport_radius
            vport_ht = vport_wd * self.vport_ht_px / float(self.vport_wd_px)

        else:
            vport_ht = 2.0 * vport_radius
            vport_wd = vport_ht * self.vport_wd_px / float(self.vport_ht_px)

        gl.glFrustum(
            -0.5 * vport_wd, 0.5 * vport_wd,    # left, right
            -0.5 * vport_ht, 0.5 * vport_ht,    # bottom, top
            self.near_plane, self.far_plane
        )

        M = Matrix4x4.look_at(self.position, self.target, self.up, False)
        gl.glMultMatrixf(M.get())

    def get_translation_speed(self, distance_from_target):
        """Returns the translation speed for ``distance_from_target``
        in units per radius.
        """
        return (distance_from_target *
                np.tan(self.fov_deg / 2.0 / 180.0 * np.pi))

    def orbit(self, x1_px, y1_px, x2_px, y2_px):
        """
        Causes the camera to "orbit" around the target point.
        This is also called "tumbling" in some software packages.
        """
        px_per_deg = self.vport_radius_px / float(self.orbit_speed)
        radians_per_px = 1.0 / px_per_deg * np.pi / 180.0

        t2p = self.position - self.target

        M = Matrix4x4.rotation_around_origin((x1_px - x2_px) * radians_per_px,
                                             self.ground)
        t2p = M * t2p
        self.up = M * self.up

        right = (self.up ^ t2p).normalized()
        M = Matrix4x4.rotation_around_origin((y1_px - y2_px) * radians_per_px,
                                             right)
        t2p = M * t2p
        self.up = M * self.up

        self.position = self.target + t2p

    def pan_delta(self, dx_px, dy_px):
        """
        This causes the scene to appear to translate right and up
        (i.e., what really happens is the camera is translated left and down).
        This is also called "panning" in some software packages.
        Passing in negative delta values causes the opposite motion.
        """
        direction = self.target - self.position
        distance_from_target = direction.length()
        direction = direction.normalized()

        speed_per_radius = self.get_translation_speed(distance_from_target)
        px_per_unit = self.vport_radius_px / speed_per_radius

        right = direction ^ self.up

        translation = (right * (-dx_px / px_per_unit) +
                       self.up * (-dy_px / px_per_unit))

        self.position = self.position + translation
        self.target = self.target + translation

    def track(self, delta_pixels, push_target=False, adj_fov=False):
        """
        This causes the camera to translate forward into the scene.
        This is also called "dollying" or "tracking" in some software packages.
        Passing in a negative delta causes the opposite motion.

        If ``push_target'' is True, the point of interest translates forward
        (or backward) *with* the camera, i.e. it's "pushed" along with the
        camera; otherwise it remains stationary.

        if ``adj_fov`` is True then the camera's FOV is adjusted to keep the
        target at the same size as before (so called "dolly zoom" or
        "trombone effect").
        """
        direction = self.target - self.position
        distance_from_target = direction.length()
        direction = direction.normalized()

        initial_ht = frustum_height_at_distance(self.fov_deg,
                                                distance_from_target)
        ## print("frustum height at distance %.4f is %.4f" % (
        ##     distance_from_target, initial_ht))

        speed_per_radius = self.get_translation_speed(distance_from_target)
        px_per_unit = self.vport_radius_px / speed_per_radius

        dolly_distance = delta_pixels / px_per_unit

        if not push_target:
            distance_from_target -= dolly_distance
            if distance_from_target < self.push_threshold * self.near_plane:
                distance_from_target = self.push_threshold * self.near_plane

        self.position += direction * dolly_distance
        self.target = self.position + direction * distance_from_target

        if adj_fov:
            # adjust FOV to match the size of the target before the dolly
            direction = self.target - self.position
            distance_from_target = direction.length()
            fov_deg = fov_for_height_and_distance(initial_ht,
                                                  distance_from_target)
            #print("fov is now %.4f" % fov_deg)
            self.fov_deg = fov_deg

    def frustum_dimensions_at_target(self, vfov_deg=None):
        if vfov_deg is None:
            vfov_deg = self.fov_deg

        direction = self.target - self.position
        distance_from_target = direction.length()

        height = frustum_height_at_distance(vfov_deg, distance_from_target)

        vp_wd, vp_ht = self.get_viewport_dimensions()
        aspect = float(vp_wd) / vp_ht

        hfov_rad = 2.0 * np.arctan(np.tan(np.radians(vfov_deg) * 0.5) * aspect)
        hfov_deg = np.degrees(hfov_rad)
        width = 2.0 * np.tan(hfov_rad * 0.5) * distance_from_target

        return (distance_from_target, hfov_deg, vfov_deg, width, height)


def frustum_height_at_distance(vfov_deg, distance):
    """Calculate the frustum height (in world units) at a given distance
    (in world units) from the camera.
    """
    height = 2.0 * distance * np.tan(np.radians(vfov_deg * 0.5))
    return height


def fov_for_height_and_distance(height, distance):
    """Calculate the FOV needed to get a given frustum height at a
    given distance.
    """
    vfov_deg = np.degrees(2.0 * np.arctan(height * 0.5 / distance))
    return vfov_deg


# END



#
# iohelper.py -- misc routines used in manipulating files, paths and urls.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os
import re
import hashlib
import mimetypes
import pathlib
import urllib.parse
import tempfile

from ginga.misc import Bunch

magic_tester = None
try:
    import magic
    have_magic = True
    # it seems there are conflicting versions of a 'magic'
    # module for python floating around...*sigh*
    if not hasattr(magic, 'from_file'):
        # TODO: do this at program start only
        magic_tester = magic.open(magic.DEFAULT_MODE)
        magic_tester.load()

except (ImportError, Exception):
    have_magic = False


# NOTE: this is a shortcut list of extensions in which we trust the
# extension over the mimetype that might be returned by the Python
# 'mimetypes' module.  You can add an extension here if the loaders
# can reliably load it and it is not recognized properly by 'mimetypes'
known_types = {
    '.fits': 'image/fits',
    '.fits.gz': 'image/fits',
    '.fits.fz': 'image/fits',
    '.asdf': 'image/asdf',
    '.jpg': 'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.tiff': 'image/tiff',
    '.gif': 'image/gif',
    '.png': 'image/png',
    '.ppm': 'image/ppm',
    '.pnm': 'image/pnm',
    '.pbm': 'image/pbm',
}


def guess_filetype(filepath):
    """Guess the type of a file."""
    # If we have python-magic, use it to determine file type
    typ = None

    # Some specific checks for known file suffixes
    _fn = filepath.lower()
    for key, mimetype in known_types.items():
        if _fn.endswith(key):
            typ = mimetype
            break

    if typ is None and have_magic:
        try:
            # it seems there are conflicting versions of a 'magic'
            # module for python floating around...*sigh*
            if hasattr(magic, 'from_file'):
                typ = magic.from_file(filepath, mime=True)

            elif magic_tester is not None:
                descrip = magic_tester.file(filepath)
                if descrip.startswith("FITS image data"):
                    return ('image', 'fits')

        except Exception as e:
            pass

    if typ is None:
        # if no magic, or magic fails, fall back to mimetypes
        try:
            typ, enc = mimetypes.guess_type(filepath)

        except Exception as e:
            # fail
            pass

    if typ:
        typ, subtyp = typ.split('/')
        return (typ, subtyp)

    raise ValueError("Can't determine file type of '%s'" % (filepath))


def get_download_path(uri, filename, cache_dir):

    # TODO: cache file by uri and timestamp
    filepath = os.path.join(cache_dir, filename)

    # find a free local copy of the file
    count, tmpfile = 1, filepath
    while os.path.exists(tmpfile):
        pfx, sfx = os.path.splitext(filepath)
        tmpfile = pfx + '_' + str(count) + sfx
        count += 1
    return tmpfile


def get_fileinfo(filespec, cache_dir=None):
    """
    Parse a file specification and return information about it.
    """
    if cache_dir is None:
        cache_dir = tempfile.gettempdir()

    # Loads first science extension by default.
    # This prevents [None] to be loaded instead.
    idx = None
    name_ext = ''

    # User specified an index using bracket notation at end of path?
    match = re.match(r'^(.+)\[(.+)\]$', filespec)
    if match:
        filespec = match.group(1)
        idx = match.group(2)
        if ',' in idx:
            hduname, extver = idx.split(',')
            hduname = hduname.strip()
            if re.match(r'^\d+$', extver):
                extver = int(extver)
                idx = (hduname, extver)
                name_ext = "[%s,%d]" % idx
            else:
                idx = idx.strip()
                name_ext = "[%s]" % idx
        else:
            if re.match(r'^\d+$', idx):
                idx = int(idx)
                name_ext = "[%d]" % idx
            else:
                idx = idx.strip()
                name_ext = "[%s]" % idx
    else:
        filespec = filespec

    url = filespec
    filepath = None

    # Does this look like a URL?
    match = re.match(r"^(\w+)://(.+)$", filespec)
    if match:
        urlinfo = urllib.parse.urlparse(filespec)
        if urlinfo.scheme == 'file':
            # local file
            filepath = str(pathlib.Path(urlinfo.path))

        else:
            path, filename = os.path.split(urlinfo.path)
            filepath = get_download_path(url, filename, cache_dir)

    else:
        # Not a URL
        filepath = filespec
        url = pathlib.Path(os.path.abspath(filepath)).as_uri()

    ondisk = os.path.exists(filepath)

    dirname, fname = os.path.split(filepath)
    fname_pfx, fname_sfx = os.path.splitext(fname)
    name = fname_pfx + name_ext

    res = Bunch.Bunch(filepath=filepath, url=url, numhdu=idx,
                      name=name, idx=name_ext, ondisk=ondisk)
    return res


def get_hdu_suffix(idx):
    if idx is None:
        return ''

    if isinstance(idx, tuple):
        assert len(idx) == 2, ValueError("idx tuple len (%d) != 2" % (
            len(idx)))
        hduname, extver = idx
        hduname = hduname.strip()
        extver = int(extver)
        if len(hduname) > 0:
            return "[%s,%d]" % (hduname, extver)
        else:
            return "[%d]" % extver

    if isinstance(idx, str):
        return "[%s]" % idx.strip()

    return "[%d]" % idx


def name_image_from_path(path, idx=None):
    (path, filename) = os.path.split(path)
    # Remove trailing .extension
    (name, ext) = os.path.splitext(filename)
    #if '.' in name:
    #    name = name[:name.rindex('.')]
    if idx is not None:
        name += get_hdu_suffix(idx)
    return name


def shorten_name(name, char_limit, side='right'):
    """Shorten `name` if it is longer than `char_limit`.
    If `side` == "right" then the right side of the name is shortened;
    if "left" then the left side is shortened.
    In either case, the suffix of the name is preserved.
    """
    # TODO: A more elegant way to do this?
    if char_limit is not None and len(name) > char_limit:
        info = get_fileinfo(name)
        if info.numhdu is not None:
            i = name.rindex('[')
            s = (name[:i], name[i:])
            len_sfx = len(s[1])
            len_pfx = char_limit - len_sfx - 4 + 1
            if len_pfx > 0:
                if side == 'right':
                    name = '{0}...{1}'.format(s[0][:len_pfx], s[1])
                elif side == 'left':
                    name = '...{0}{1}'.format(s[0][-len_pfx:], s[1])
            else:
                name = '...{0}'.format(s[1])
        else:
            len1 = char_limit - 3 + 1
            if side == 'right':
                name = '{0}...'.format(name[:len1])
            elif side == 'left':
                name = '...{0}'.format(name[-len1:])

    return name


def gethex(s):
    return hashlib.sha1(s.encode()).hexdigest()


def get_thumbpath(thumbdir, path, makedir=True):
    if path is None:
        return None

    path = os.path.abspath(path)
    dirpath, filename = os.path.split(path)

    if not os.path.exists(thumbdir):
        if not makedir:
            raise ValueError("Thumb directory does not exist: %s" % (
                thumbdir))

        try:
            os.makedirs(thumbdir)
            # Write meta file
            metafile = os.path.join(thumbdir, "meta")
            with open(metafile, 'w') as out_f:
                out_f.write("srcdir: %s\n" % (dirpath))

        except OSError as e:
            raise Exception("Could not make thumb directory '%s': %s" % (
                thumbdir, str(e)))

    # Get location of thumb
    modtime = os.stat(path).st_mtime
    thumb_fname = gethex("%s.%s" % (filename, modtime))
    thumbpath = os.path.join(thumbdir, thumb_fname + ".jpg")
    return thumbpath
